from typing import Dict, Optional, Tuple
import functools

from torch.cuda.amp import custom_bwd, custom_fwd
import torch
import torch.distributed

from realhf.impl.model.parallelism.model_parallel.utils import VocabUtility
from realhf.impl.model.utils.functional import build_leave_one_indices
import realhf.base.constants as constants


class KLController:

    def __init__(self, kl_coef):
        self.value = kl_coef

    def update(self):
        raise NotImplementedError()


class AdaptiveKLController(KLController):
    """
    Adaptive KL controller described in the paper:
    https://arxiv.org/pdf/1909.08593.pdf
    """

    def __init__(self, init_kl_coef, target, horizon):
        self.value = init_kl_coef
        self.target = target
        self.horizon = horizon

    def update(self, current, n_steps):
        target = self.target
        proportional_error = torch.clip(current / target - 1, -0.2, 0.2)
        mult = 1 + proportional_error * n_steps / self.horizon
        self.value = self.value * mult


class FixedKLController(KLController):
    """Fixed KL controller."""

    def __init__(self, kl_coef):
        self.value = kl_coef

    def update(self, current, n_steps):
        pass


def actor_loss_fn(
    logprobs: torch.FloatTensor,
    old_logprobs: torch.FloatTensor,
    advantages: torch.FloatTensor,
    eps_clip: float,
    loss_mask: Optional[torch.BoolTensor] = None,
) -> Tuple[torch.Tensor, Dict]:
    """Compute PPO actor loss function.

    There is no shape requirements for the inputs, but they must have the same shape.
    Either [bs, max_seqlen] for batch padded inputs or [tot_seqlen] for padded inputs.

    Args:
        logprobs (torch.FloatTensor): Log probabilities of actions.
        old_logprobs (torch.FloatTensor): Old log probabilities of actions.
        advantages (torch.FloatTensor): GAE (normalized) advantages.
        eps_clip (float): Clip ratio of PPO.
        loss_mask (Optional[torch.BoolTensor], optional): Mask for loss computation.
            1 if valid else 0. Defaults to None.

    Returns:
        Tuple[torch.Tensor, Dict]: Scalar loss and statistics.
    """
    assert logprobs.dtype == torch.float32
    assert old_logprobs.dtype == torch.float32
    assert advantages.dtype == torch.float32

    # clone inference tensors
    if old_logprobs.is_inference():
        old_logprobs = old_logprobs.clone()
    if advantages.is_inference():
        advantages = advantages.clone()

    if loss_mask is not None:
        loss_mask_count = loss_mask.count_nonzero()
        # For numerical stability.
        ratio = torch.where(loss_mask, torch.exp(logprobs - old_logprobs), 0)
        approx_kl = torch.where(
            loss_mask, (logprobs - old_logprobs).detach(), 0.0
        )
    else:
        ratio = torch.exp(logprobs - old_logprobs)
        approx_kl = (logprobs - old_logprobs).detach()

    clipped_ratio = torch.clamp(ratio, 1.0 - eps_clip, 1.0 + eps_clip)
    pg_loss1 = -advantages * ratio
    pg_loss2 = -advantages * clipped_ratio

    if loss_mask is not None:
        pg_loss = (
            torch.where(loss_mask, torch.max(pg_loss1, pg_loss2), 0).sum()
            / loss_mask_count
        )
    else:
        pg_loss = torch.max(pg_loss1, pg_loss2).mean()

    clip_mask = pg_loss1.detach() < pg_loss2.detach()
    if loss_mask is not None:
        proportion_clipped = (
            clip_mask.logical_and_(loss_mask).count_nonzero() / loss_mask_count
        )
        importance_weight = (
            torch.where(loss_mask, ratio.detach(), 0).sum() / loss_mask_count
        )
        approx_kl = approx_kl.sum() / loss_mask_count
    else:
        proportion_clipped = clip_mask.count_nonzero()
        importance_weight = ratio.detach().mean()
        approx_kl = approx_kl.mean()
    # Remain torch.CudaTensor here for all-reduce after train step.
    stat = dict(
        clip_ratio=proportion_clipped,
        importance_weight=importance_weight,
        approx_kl=approx_kl,
    )

    return pg_loss, stat


class _VocabParallelMemoryEfficientPPOLoss(torch.autograd.Function):

    @staticmethod
    def forward(
        ctx,
        vocab_parallel_logits,
        cu_seqlens,
        packed_input_ids,
        ppo_loss_mask,
        old_logprobs,
        advantages,
        eps_clip,
    ):

        target = labels = torch.nn.functional.pad(
            packed_input_ids[1:], (0, 1), value=0.0
        )
        # Maximum value along vocab dimension across all GPUs.
        logits_max = torch.max(vocab_parallel_logits, dim=-1)[0]
        torch.distributed.all_reduce(
            logits_max,
            op=torch.distributed.ReduceOp.MAX,
            group=constants.model_parallel_group(),
        )
        # Subtract the maximum value.
        vocab_parallel_logits = vocab_parallel_logits - logits_max.unsqueeze(
            dim=-1
        )

        # Get the partition's vocab indecies
        get_vocab_range = VocabUtility.vocab_range_from_per_partition_vocab_size
        partition_vocab_size = vocab_parallel_logits.size()[-1]
        rank = constants.model_parallel_rank()
        world_size = constants.model_parallel_world_size()
        vocab_start_index, vocab_end_index = get_vocab_range(
            partition_vocab_size, rank, world_size
        )

        # Create a mask of valid vocab ids (1 means it needs to be masked).
        target_mask = (target < vocab_start_index) | (target >= vocab_end_index)
        masked_target = target.clone() - vocab_start_index
        masked_target[target_mask] = 0

        # Get predicted-logits = logits[target].
        # For Simplicity, we convert logits to a 2-D tensor with size
        # [*, partition-vocab-size] and target to a 1-D tensor of size [*].
        logits_2d = vocab_parallel_logits.view(-1, partition_vocab_size)
        masked_target_1d = masked_target.view(-1)
        arange_1d = torch.arange(
            start=0, end=logits_2d.size()[0], device=logits_2d.device
        )
        predicted_logits_1d = logits_2d[arange_1d, masked_target_1d]
        predicted_logits_1d = predicted_logits_1d.clone().contiguous()
        predicted_logits = predicted_logits_1d.view_as(target)
        predicted_logits[target_mask] = 0.0
        # All reduce is needed to get the chunks from other GPUs.
        torch.distributed.all_reduce(
            predicted_logits,
            op=torch.distributed.ReduceOp.SUM,
            group=constants.model_parallel_group(),
        )

        # Sum of exponential of logits along vocab dimension across all GPUs.
        exp_logits = vocab_parallel_logits
        torch.exp(vocab_parallel_logits, out=exp_logits)
        sum_exp_logits = exp_logits.sum(dim=-1)
        torch.distributed.all_reduce(
            sum_exp_logits,
            op=torch.distributed.ReduceOp.SUM,
            group=constants.model_parallel_group(),
        )

        # Loss = log(sum(exp(logits))) - predicted-logit.
        new_logprobs = predicted_logits - torch.log(sum_exp_logits)

        # Normalize and optionally smooth logits
        exp_logits.div_(sum_exp_logits.unsqueeze(dim=-1))

        leave_one_indices = build_leave_one_indices(new_logprobs, cu_seqlens)
        new_logprobs = new_logprobs[leave_one_indices] * ppo_loss_mask
        new_logprobs = new_logprobs.float()

        # For numerical stability.
        ratio = torch.where(
            ppo_loss_mask, torch.exp(new_logprobs - old_logprobs), 0
        )

        loss_mask_count = ppo_loss_mask.count_nonzero()
        approx_kl = (
            torch.where(
                ppo_loss_mask, (old_logprobs - new_logprobs).detach(), 0.0
            ).sum()
            / loss_mask_count
        )
        importance_weight = (
            torch.where(ppo_loss_mask, ratio.detach(), 0).sum()
            / loss_mask_count
        )

        clipped_ratio = torch.clamp(ratio, 1.0 - eps_clip, 1.0 + eps_clip)
        pg_loss1 = -advantages * ratio
        pg_loss2 = -advantages * clipped_ratio

        clip_mask = (pg_loss1 < pg_loss2).detach()
        proportion_clipped = (
            clip_mask.logical_and_(ppo_loss_mask).count_nonzero()
            / loss_mask_count
        )

        # Store softmax, target-mask and masked-target for backward pass.
        ctx.save_for_backward(
            exp_logits,
            target_mask,
            masked_target_1d,
            leave_one_indices,
            ppo_loss_mask,
            ratio,
            advantages,
        )
        ctx.eps_clip = eps_clip

        return (
            torch.max(pg_loss1, pg_loss2),
            importance_weight,
            proportion_clipped,
            approx_kl,
        )

    @staticmethod
    def backward(ctx, grad_output, g1, g2, g3):
        # Retreive tensors from the forward path.
        (
            softmax,
            target_mask,
            masked_target_1d,
            leave_one_indices,
            ppo_loss_mask,
            ratio,
            advantages,
        ) = ctx.saved_tensors
        eps_clip = ctx.eps_clip

        # ppo backward
        clipped_ratio = torch.clamp(ratio, 1.0 - eps_clip, 1.0 + eps_clip)
        pg_loss1 = -advantages * ratio
        pg_loss2 = -advantages * clipped_ratio
        pg_loss1_larger_mask = pg_loss1 > pg_loss2
        unclip_mask = (ratio >= 1.0 - eps_clip) & (ratio <= 1.0 + eps_clip)

        g_ = -grad_output * advantages
        grad_ratio = torch.where(pg_loss1_larger_mask & unclip_mask, g_, 0.0)
        grad_new_logp = torch.where(ppo_loss_mask, grad_ratio * ratio, 0.0)
        _grad_new_logp = grad_new_logp.new_zeros(
            softmax.shape[0], dtype=torch.float16
        )
        _grad_new_logp[leave_one_indices] = grad_new_logp.half()

        # All the inputs have softmax as thier gradient.
        grad_input = softmax
        # For simplicity, work with the 2D gradient.
        partition_vocab_size = softmax.size()[-1]
        grad_2d = grad_input.view(-1, partition_vocab_size)

        # Add the gradient from matching classes.
        arange_1d = torch.arange(
            start=0, end=grad_2d.size()[0], device=grad_2d.device
        )

        softmax_update = 1.0 - target_mask.view(-1).float()
        grad_2d[arange_1d, masked_target_1d] -= softmax_update

        # Finally elementwise multiplication with the output gradients.
        grad_input.mul_(_grad_new_logp.unsqueeze(dim=-1))

        return grad_input, None, None, None, None, None, None


class _MemoryEfficientPPOActorLossFn(torch.autograd.Function):

    @staticmethod
    @custom_fwd
    def forward(
        ctx,
        logits,
        cu_seqlens,
        packed_input_ids,
        ppo_loss_mask,
        old_logprobs,
        advantages,
        eps_clip,
    ):
        labels = torch.nn.functional.pad(
            packed_input_ids[1:], (0, 1), value=0.0
        )
        _new_logprobs = torch.nn.functional.log_softmax(logits, dim=-1)
        new_logprobs_labels = _new_logprobs.gather(
            dim=-1, index=labels.unsqueeze(-1)
        ).squeeze(-1)
        leave_one_indices = build_leave_one_indices(logits, cu_seqlens)
        new_logprobs = torch.where(
            ppo_loss_mask, new_logprobs_labels[leave_one_indices], 0.0
        )
        new_logprobs = new_logprobs.float()

        # For numerical stability.
        ratio = torch.where(
            ppo_loss_mask, torch.exp(new_logprobs - old_logprobs), 0
        )

        loss_mask_count = ppo_loss_mask.count_nonzero()
        approx_kl = (
            torch.where(
                ppo_loss_mask, (old_logprobs - new_logprobs).detach(), 0.0
            ).sum()
            / loss_mask_count
        )
        importance_weight = (
            torch.where(ppo_loss_mask, ratio.detach(), 0).sum()
            / loss_mask_count
        )

        clipped_ratio = torch.clamp(ratio, 1.0 - eps_clip, 1.0 + eps_clip)
        pg_loss1 = -advantages * ratio
        pg_loss2 = -advantages * clipped_ratio

        clip_mask = (pg_loss1 < pg_loss2).detach()
        proportion_clipped = (
            clip_mask.logical_and_(ppo_loss_mask).count_nonzero()
            / loss_mask_count
        )

        ctx.save_for_backward(
            logits, leave_one_indices, labels, ppo_loss_mask, ratio, advantages
        )
        ctx.eps_clip = eps_clip

        return (
            torch.max(pg_loss1, pg_loss2),
            importance_weight,
            proportion_clipped,
            approx_kl,
        )

    @staticmethod
    @custom_bwd
    def backward(ctx, grad_output, g1, g2, g3):
        logits, leave_one_indices, labels, ppo_loss_mask, ratio, advantages = (
            ctx.saved_tensors
        )
        eps_clip = ctx.eps_clip

        clipped_ratio = torch.clamp(ratio, 1.0 - eps_clip, 1.0 + eps_clip)
        pg_loss1 = -advantages * ratio
        pg_loss2 = -advantages * clipped_ratio
        pg_loss1_larger_mask = pg_loss1 > pg_loss2
        unclip_mask = (ratio >= 1.0 - eps_clip) & (ratio <= 1.0 + eps_clip)

        g_ = -grad_output * advantages
        grad_ratio = torch.where(pg_loss1_larger_mask & unclip_mask, g_, 0.0)
        grad_new_logp = torch.where(ppo_loss_mask, grad_ratio * ratio, 0.0)
        _grad_new_logp = grad_new_logp.new_zeros(
            logits.shape[0], dtype=torch.float16
        )
        _grad_new_logp[leave_one_indices] = grad_new_logp.half()

        grad_logits = torch.nn.functional.softmax(logits, dim=-1)
        grad_logits.scatter_add_(
            1, labels.view(-1, 1), torch.full_like(grad_logits, fill_value=-1)
        )
        grad_logits.mul_(_grad_new_logp.unsqueeze(1))

        return grad_logits, None, None, None, None, None, None, None


def memory_efficient_ppo_loss_fn(
    logits,
    cu_seqlens,
    packed_input_ids,
    ppo_loss_mask,
    old_logprobs,
    advantages,
    eps_clip,
):
    if constants.model_parallel_world_size() == 1:
        return _MemoryEfficientPPOActorLossFn.apply(
            logits,
            cu_seqlens,
            packed_input_ids,
            ppo_loss_mask,
            old_logprobs,
            advantages,
            eps_clip,
        )
    else:
        return _VocabParallelMemoryEfficientPPOLoss.apply(
            logits,
            cu_seqlens,
            packed_input_ids,
            ppo_loss_mask,
            old_logprobs,
            advantages,
            eps_clip,
        )


def _huber_loss(x: torch.Tensor, y: torch.Tensor, delta: float):
    diff = torch.abs(x - y)
    return torch.where(
        diff < delta, 0.5 * diff**2, delta * (diff - 0.5 * delta)
    )


def _mse_loss(x: torch.Tensor, y: torch.Tensor):
    return 0.5 * (x - y) ** 2


def critic_loss_fn(
    value: torch.FloatTensor,
    old_value: torch.FloatTensor,
    target_value: torch.FloatTensor,
    value_eps_clip: float,
    loss_mask: Optional[torch.FloatTensor] = None,
    loss_fn_type: str = "mse",
) -> Tuple[torch.Tensor, Dict]:
    """Compute PPO critic loss function given padded batch inputs.

    There is no shape requirements for the inputs, but they must have the same shape.
    Either [bs, max_seqlen] for batch padded inputs or [tot_seqlen] for padded inputs.

    Args:
        value (torch.FloatTensor): Values. The position of the final token is not included.
            (The whole generated sequence is not a state.)
        old_value (torch.FloatTensor): Old values.
        target_value (torch.FloatTensor): Returns computed by GAE.
        value_eps_clip (float): Clip ratio.
        loss_mask (Optional[torch.FloatTensor], optional): Mask for loss computation.
            1 if valid else 0. Defaults to None.
        loss_fn_type (str, optional): Type of loss function. Defaults to 'huber'.

    Returns:
        Tuple[torch.Tensor, Dict]: Scalar loss and statistics.
    """
    assert value.dtype == torch.float32
    assert old_value.dtype == torch.float32
    assert target_value.dtype == torch.float32

    if loss_fn_type == "huber":
        loss_fn = functools.partial(_huber_loss, delta=10.0)
    elif loss_fn_type == "mse":
        loss_fn = _mse_loss
    else:
        raise NotImplementedError(f"Unknown loss fn type: {loss_fn_type}")

    if target_value.is_inference():
        target_value = target_value.clone()  # clone a inference tensor

    value_loss_original = loss_fn(value, target_value)

    value_clipped = old_value + (value - old_value).clamp(
        -value_eps_clip, value_eps_clip
    )

    value_loss_clipped = loss_fn(value_clipped, target_value)

    value_loss = torch.max(value_loss_original, value_loss_clipped)

    with torch.no_grad():
        clip_mask = value_loss_clipped.detach() > value_loss_original.detach()
        if loss_mask is not None:
            mask_count = loss_mask.count_nonzero()
            proportion_clipped = (
                clip_mask.logical_and_(loss_mask).count_nonzero() / mask_count
            )
        else:
            proportion_clipped = clip_mask.count_nonzero()

        stat = dict(clip_ratio=proportion_clipped)

    if loss_mask is not None:
        value_loss = torch.where(loss_mask, value_loss, 0).sum() / mask_count
    else:
        value_loss = value_loss.mean()

    return value_loss, stat


@torch.no_grad()
def compute_rewards(
    kl_ctl: float,
    clip_reward_value: float,
    log_probs: torch.FloatTensor,
    ref_log_probs: torch.FloatTensor,
    reward_score: torch.FloatTensor,
    eos_indices: torch.LongTensor,
    seq_no_eos_mask: torch.FloatTensor,
) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
    """Compute rewards for padded batch inputs.

    Args:
        kl_ctl (float): Coefficient of KL rewards.
        clip_reward_value (float): Clip value of rewards. Preventing extreme reward numbers.
        log_probs (torch.FloatTensor): Log probabilities of actions. Shape [bs, max_seqlen - 1].
        ref_log_probs (torch.FloatTensor): Log probabilities outputed by reference models. Shape [bs, max_seqlen - 1].
        reward_score (torch.FloatTensor): Scores outputed by the reward model. Shape [bs].
        eos_indices (torch.LongTensor): Indices of EOS tokens. Shape [bs].
            Used for adding score rewards onto KL rewards.
        seq_no_eos_mask (torch.FloatTensor): Indicator of no EOS tokens in a sequence. Shape [bs].
            1 if no EOS else 0.

    Returns:
        Tuple[torch.FloatTensor, torch.FloatTensor]: Pure KL rewards and total rewards.
            Both of shape [bs, max_seqlen - 1].
    """
    kl_rewards = -kl_ctl * (log_probs - ref_log_probs)
    for i in range(kl_rewards.shape[0]):
        # Set KL rewards *at EOS* and after EOS to 0.
        # The final *state* is the sequence without EOS, so the final KL reward is assigned to this state.
        # The next "environment step" indicates a "done" by outputting an EOS token, therefore no rewards afterwards.
        kl_rewards[i, eos_indices[i] :] = 0.0

    reward_clip = torch.clamp(
        reward_score, -clip_reward_value, clip_reward_value
    )
    score_rewards = torch.zeros_like(kl_rewards)
    # This is assigned to the token before EOS, which rewards the output of the EOS token.
    score_rewards.scatter_(
        -1, (eos_indices - 1).unsqueeze(-1), reward_clip.unsqueeze(-1)
    )
    score_rewards = score_rewards * (
        1 - seq_no_eos_mask.unsqueeze(1)
    )  # only compute final rewards with EOS
    return kl_rewards, kl_rewards + score_rewards


@torch.no_grad()
def get_advantages_and_returns(
    gamma: float,
    lam: float,
    values: torch.FloatTensor,
    rewards: torch.FloatTensor,
    seq_no_eos_mask: torch.FloatTensor,
) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
    """Compute GAE and returns given padded batch inputs.

    Adopted from https://github.com/CarperAI/trlx/blob/main/trlx/models/modeling_ppo.py#L134

    Args:
        gamma (float): Discount factor.
        lam (float): GAE discount factor.
        values (torch.FloatTensor): Values of shape [bs, max_seqlen] (with bootstrapped values).
        rewards (torch.FloatTensor): Rewards of shape [bs, max_seqlen - 1].
        seq_no_eos_mask (torch.FloatTensor): Indicator of no EOS tokens in a sequence. Shape [bs].
            1 if no EOS else 0. Used for masking out bootstrap values for terminated sequences.

    Returns:
        Tuple[torch.FloatTensor, torch.FloatTensor]: GAE and returns of shape [bs, max_seqlen - 1].
    """
    assert values.shape[1] == rewards.shape[1] + 1
    lastgaelam = 0
    advantages_reversed = []
    length = rewards.size()[-1]
    for t in reversed(range(length)):
        nextvalues = values[:, t + 1]
        if t == length - 1:
            nextvalues *= seq_no_eos_mask
        delta = rewards[:, t] + gamma * nextvalues - values[:, t]
        lastgaelam = delta + gamma * lam * lastgaelam
        advantages_reversed.append(lastgaelam)
    advantages = torch.stack(advantages_reversed[::-1], dim=1)
    returns = advantages + values[:, :-1]
    return advantages, returns


@torch.no_grad()
def get_packed_rewards(
    kl_ctl: float,
    clip_reward_value: float,
    log_probs: torch.FloatTensor,
    ref_log_probs: torch.FloatTensor,
    reward_score: torch.FloatTensor,
    short1cu_seqlens: torch.IntTensor,
    seq_no_eos_mask: torch.BoolTensor,
) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
    # Here log_probs/ref_log_probs is one-step shorter than packed_input_ids (the last step is removed),
    # so the log_probs at the EOS token is not included in this tensor.
    # We directly add reward scores of each sequence onto the final token of each sequence.
    tot_rewards = -kl_ctl * (log_probs - ref_log_probs)
    kl_rewards = tot_rewards.clone()
    reward_score = reward_score.clip(-clip_reward_value, clip_reward_value)
    tot_rewards[short1cu_seqlens[1:] - 1] += torch.where(
        seq_no_eos_mask, 0, reward_score
    )
    return kl_rewards, tot_rewards


def pygae2d_olp(
    rewards: torch.FloatTensor,
    values: torch.FloatTensor,
    dones: torch.BoolTensor,
    truncates: torch.BoolTensor,
    gamma: float,
    lam: float,
) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
    episode_length = int(rewards.shape[1])
    masks = 1 - dones.float()
    truncate_mask = 1 - truncates.float()
    delta = rewards + gamma * values[:, 1:] * masks[:, 1:] - values[:, :-1]
    adv = torch.zeros_like(rewards)
    gae = torch.zeros_like(rewards[:, 0])
    m = gamma * lam * masks[:, 1:]
    step = episode_length - 1
    while step >= 0:
        # if env is terminated compulsively, then abandon the finnal step
        # i.e. advantage of final step is 0, values target of final step is predicted values
        gae = (delta[:, step] + m[:, step] * gae) * truncate_mask[:, step + 1]
        adv[:, step] = gae
        step -= 1
    return adv, adv + values[:, :-1]


def pygae1d_nolp_misalign(
    rewards: torch.FloatTensor,
    values: torch.FloatTensor,
    cu_seqlens_: torch.IntTensor,
    bootstrap: torch.FloatTensor,
    gamma: float,
    lam: float,
) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
    cu_seqlens = cu_seqlens_.clone()
    cu_seqlens[1:] += torch.ones_like(cu_seqlens_[1:]).cumsum(0)

    bs = cu_seqlens_.shape[0] - 1
    assert values.shape[0] == rewards.shape[0] + bs
    advantages_reversed = []
    returns_reversed = []
    for i in reversed(range(bs)):
        v_offset = cu_seqlens[i]
        r_offset, r_end = cu_seqlens_[i], cu_seqlens_[i + 1]
        assert cu_seqlens[i + 1] - v_offset - 1 == r_end - r_offset
        lastgaelam = 0
        for t in reversed(range(r_end - r_offset)):
            nextvalues = values[v_offset + t + 1]
            if t == r_end - r_offset - 1:
                nextvalues *= bootstrap[i]
            delta = (
                rewards[r_offset + t]
                + gamma * nextvalues
                - values[v_offset + t]
            )
            lastgaelam = delta + gamma * lam * lastgaelam
            advantages_reversed.append(lastgaelam)
            returns_reversed.append(lastgaelam + values[v_offset + t])

    advantages = torch.stack(advantages_reversed[::-1])
    returns = torch.stack(returns_reversed[::-1])
    return advantages, returns


def pygae2d_nolp(
    rewards: torch.FloatTensor,
    values: torch.FloatTensor,
    on_reset: torch.BoolTensor,
    truncates: torch.BoolTensor,
    gamma: float,
    lam: float,
) -> torch.FloatTensor:
    on_reset = on_reset.float()
    truncates = truncates.float()
    episode_length = int(rewards.shape[1])
    delta = (
        rewards + gamma * values[:, 1:] * (1 - on_reset[:, 1:]) - values[:, :-1]
    )

    gae = torch.zeros_like(rewards[:, 0])
    adv = torch.zeros_like(rewards)

    # 1. If the next step is a new episode, GAE doesn't propagate back
    # 2. If the next step is a truncated final step, the backpropagated GAE is -V(t),
    #    which is not correct. We ignore it such that the current GAE is r(t-1)+ɣV(t)-V(t-1)
    # 3. If the next step is a done final step, the backpropagated GAE is zero.
    m = gamma * lam * (1 - on_reset[:, 1:]) * (1 - truncates[:, 1:])

    step = episode_length - 1
    while step >= 0:
        gae = delta[:, step] + m[:, step] * gae
        adv[:, step] = gae
        step -= 1

    return adv, adv + values[:, :-1]


def cugae1d_nolp_misalign_func(
    rewards: torch.FloatTensor,
    values: torch.FloatTensor,
    cu_seqlens: torch.IntTensor,
    truncate: torch.BoolTensor,
    gamma: float,
    lam: float,
) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
    """Compute GAE over a batch of packed sequences with different lengths.

    This function assumes that rewards and values are packed into an 1D tensor.
    Values are longer than rewards by the number of sequences in rewards because of bootstrapping.
    cu_seqlens marks the bounary of sequences in rewards.

    The final step of each sequence is *NOT* overlapped with the first step of the next sequence,
    and rewards/values do not have the same length, so this function is suffixed with
    "nolp" (non-overlap) and "misalign".

    Args:
        rewards (torch.FloatTensor): Shape [total_seqlen], rewards across sequences.
        values (torch.FloatTensor): Shape [total_seqlen + batch_size], values across sequences.
            Values are bootstrapped, so it's longer than rewards.
        cu_seqlens (torch.IntTensor): Marker of sequence boundaries in rewards,
            e.g., [0, s1, s1+s2, ..., total_seqlen]. It should starts with 0 and ends with total_seqlen.
        truncate (torch.BoolTensor): Whether each sequence is truncated because of exceeding max length.
            If truncate, the next value of the last step will be bootstraped, otherwise 0.
        gamma (float): Discount factor.
        lam (float): GAE discount factor.

    Returns:
        Tuple[torch.FloatTensor, torch.FloatTensor]: Advantages and returns (value targets).
            Both have the same shape as rewards.
    """
    import realhf._C.cugae as gae_cuda

    assert len(rewards.shape) == len(values.shape) == len(cu_seqlens.shape) == 1
    assert cu_seqlens[0] == 0 and cu_seqlens[-1] == rewards.shape[0]
    return gae_cuda.gae_1d_nolp_misalign(
        rewards, values, cu_seqlens, truncate, gamma, lam
    )


def cugae2d_olp_func(
    rewards: torch.FloatTensor,
    values: torch.FloatTensor,
    dones: torch.BoolTensor,
    truncates: torch.BoolTensor,
    gamma: float,
    lam: float,
) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
    """Compute GAE over batched sequences with variable lengths, assuming overlapped sequences.

    This function assumes that rewards and values are batched as 2D tensors.
    The first dimension is batch_size and the second dimension is the number of collected timesteps.
    Each batch slot may contain multiple sequences, and sequences may have different lengths.
    The length of each sequence is marked by dones.

    `dones` marks the termination of each sequence, no matter it's truncated or not.
    `truncates` marks truncation and its nonzero indices must be the subset of `dones`.
    If truncate, abandon GAE computation of the last step (because we don't have the bootstrapped
    value in this case) and start from the second last step.

    The final step of each sequence *is overlapped* by the first step of the next sequence,
    i.e., auto-reset, which has widely used in libraries such as gym. In other words, the
    steps where `dones` is True are actually the first steps of sequences. Therefore,
    this function is suffixed with "olp" (overlap).

    Args:
        rewards (torch.FloatTensor): Shape [batch_size, seqlen].
        values (torch.FloatTensor): Shape [batch_size, seqlen + 1], with one more bootstrap step.
        dones (torch.BoolTensor): Shape [batch_size, seqlen + 1].
        truncates (torch.BoolTensor): Shape [batch_size, seqlen + 1].
        gamma (float): Discount factor.
        lam (float): GAE discount factor.

    Returns:
        Tuple[torch.FloatTensor, torch.FloatTensor]: Advantages and returns (value targets).
            Both have the same shape as rewards.
    """
    import realhf._C.cugae as gae_cuda

    truncates_indices = truncates.nonzero()
    assert torch.all(dones[truncates_indices[:, 0], truncates_indices[:, 1]])
    done_indices = dones.nonzero()
    num_dones = dones.float().sum(1)
    max_num_dones = int(num_dones.max())
    cu_num_dones = torch.nn.functional.pad(
        num_dones.cumsum(0), (1, 0), value=0
    ).int()
    is_truncate = truncates[done_indices[:, 0], done_indices[:, 1]]
    return gae_cuda.gae_2d_olp(
        rewards,
        values,
        dones,
        done_indices[:, 1].int(),
        cu_num_dones,
        max_num_dones,
        is_truncate,
        gamma,
        lam,
    )


def cugae2d_nolp_func(
    rewards: torch.FloatTensor,
    values: torch.FloatTensor,
    on_reset: torch.BoolTensor,
    truncates: torch.BoolTensor,
    gamma: float,
    lam: float,
) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
    """Compute GAE over batched sequences with variable lengths, assuming non-overlapped sequences.

    This function assumes that rewards and values are batched as 2D tensors.
    The first dimension is batch_size and the second dimension is the number of collected timesteps.
    Each batch slot may contain multiple sequences, and sequences may have different lengths.
    The length of each sequence is marked by `on_reset`.

    `on_reset` marks the beginning of each sequence. `truncates` marks truncation.
    If truncate, values will be bootstrapped from the `done` step.

    The final step of each sequence is *NOT* overlapped by the first step of the next sequence.
    Each sequence will be complete. The last step should only have observations but no rewards.
    This is used in SRL. Therefore, this function is suffixed with "nolp" (non-overlap).

    Args:
        rewards (torch.FloatTensor): Shape [batch_size, seqlen].
        values (torch.FloatTensor): Shape [batch_size, seqlen + 1], with one more bootstrap step.
        dones (torch.BoolTensor): Shape [batch_size, seqlen + 1].
        truncates (torch.BoolTensor): Shape [batch_size, seqlen + 1].
        gamma (float): Discount factor.
        lam (float): GAE discount factor.

    Returns:
        Tuple[torch.FloatTensor, torch.FloatTensor]: Advantages and returns (value targets).
            Both have the same shape as rewards.
    """
    import realhf._C.cugae as gae_cuda

    dones = on_reset[:, 1:]
    truncates_indices = truncates[:, :-1].nonzero()
    assert torch.all(dones[truncates_indices[:, 0], truncates_indices[:, 1]])
    on_reset_indices = on_reset.nonzero()
    num_resets = on_reset.float().sum(1)
    max_num_resets = int(num_resets.max())
    cu_num_resets = torch.nn.functional.pad(
        num_resets.cumsum(0), (1, 0), value=0
    ).int()
    truncates = torch.cat(
        [torch.zeros_like(truncates[:, 0:1]), truncates[:, :-1]], dim=1
    )
    bootstrap = truncates[on_reset_indices[:, 0], on_reset_indices[:, 1]]
    return gae_cuda.gae_2d_nolp(
        rewards,
        values,
        on_reset,
        on_reset_indices[:, 1].int(),
        cu_num_resets,
        max_num_resets,
        bootstrap,
        gamma,
        lam,
    )


@torch.no_grad()
def get_packed_advantages_and_returns(
    gamma: float,
    lam: float,
    values: torch.FloatTensor,
    rewards: torch.FloatTensor,
    short1cu_seqlens: torch.IntTensor,
    seq_no_eos_mask: torch.FloatTensor,
) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
    try:
        return cugae1d_nolp_misalign_func(
            rewards,
            values,
            short1cu_seqlens.int(),
            seq_no_eos_mask.bool(),
            gamma,
            lam,
        )
    except ModuleNotFoundError:
        return pygae1d_nolp_misalign(
            rewards, values, short1cu_seqlens, seq_no_eos_mask, gamma, lam
        )



# This file standardizes the name-resolve names used by different components of the system.
import getpass

USER_NAMESPACE = getpass.getuser()


def registry_root(user):
    return f"trial_registry/{user}"


def trial_registry(experiment_name, trial_name):
    return f"trial_registry/{USER_NAMESPACE}/{experiment_name}/{trial_name}"


def trial_root(experiment_name, trial_name):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}"


def worker_status(experiment_name, trial_name, worker_name):
    return (
        f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/status/{worker_name}"
    )


def worker_root(experiment_name, trial_name):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/worker/"


def worker(experiment_name, trial_name, worker_name):
    return (
        f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/worker/{worker_name}"
    )


def worker2(experiment_name, trial_name, worker_type, worker_index):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/worker/{worker_type}/{worker_index}"


def inference_stream(experiment_name, trial_name, stream_name):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/inference_stream/{stream_name}"


def inference_stream_constant(
    experiment_name, trial_name, stream_name, constant_name
):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/inference_stream_consts/{stream_name}/{constant_name}"


def sample_stream(experiment_name, trial_name, stream_name):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/sample_stream/{stream_name}"


def request_reply_stream(experiment_name, trial_name, stream_name):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/request_reply_stream/{stream_name}"


def trainer_ddp_peer(experiment_name, trial_name, model_name):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/trainer_ddp_peer/{model_name}"


def trainer_ddp_local_peer(experiment_name, trial_name, host_name, model_name):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/trainer_ddp_local_peer/{host_name}/{model_name}"


def trainer_ddp_master(experiment_name, trial_name, model_name):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/trainer_ddp_master/{model_name}"


def worker_key(experiment_name, trial_name, key):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/worker_key/{key}"


def parameter_subscription(experiment_name, trial_name):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/parameter_sub"


def parameter_server(experiment_name, trial_name, parameter_id_str):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/parameter_server/{parameter_id_str}"


def shared_memory(experiment_name, trial_name, stream_name):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/shared_memory/{stream_name}"


def shared_memory_dock_server(
    experiment_name, trial_name, stream_name, server_type
):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/shared_memory_dock_server/{server_type}/{stream_name}"


def ray_cluster(experiment_name, trial_name, name):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/ray_cluster/{name}"


def model_controller(experiment_name, trial_name, model_name, dp_rank, mp_rank):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/model_controller/{model_name}/{dp_rank}/{mp_rank}"


def model_controller_barrier(
    experiment_name, trial_name, model_name, dp_rank, mp_rank
):
    return f"{USER_NAMESPACE}/{experiment_name}/{trial_name}/model_controller_barrier/{model_name}/{dp_rank}/{mp_rank}"



import argparse
import datetime
import getpass
import sys

import hydra

# NOTE: Register all implemented experiments inside ReaL.
from realhf.api.quickstart.entrypoint import QUICKSTART_FN
from realhf.base.cluster import spec as cluster_spec
import realhf.experiments.common.dpo_exp
import realhf.experiments.common.ppo_exp
import realhf.experiments.common.rw_exp
import realhf.experiments.common.sft_exp


def main():
    parser = argparse.ArgumentParser(prog="ReaL Quickstart")
    subparsers = parser.add_subparsers(dest="cmd", help="sub-command help")
    subparsers.required = True
    for k, v in QUICKSTART_FN.items():
        subparser = subparsers.add_parser(k)
        subparser.set_defaults(func=v)
    args = parser.parse_known_args()[0]

    launch_hydra_task(args.cmd, QUICKSTART_FN[args.cmd])


def launch_hydra_task(name: str, func: hydra.TaskFunction):
    # Disable hydra logging.
    if not any("hydra/job_logging=disabled" in x for x in sys.argv):
        sys.argv += ["hydra/job_logging=disabled"]

    if any("experiment_name=" in x for x in sys.argv):
        experiment_name = next(
            x for x in sys.argv if "experiment_name=" in x
        ).split("=")[1]
        if "_" in experiment_name:
            raise RuntimeError("experiment_name should not contain `_`.")
    else:
        experiment_name = f"quickstart-{name}"
        print(
            f"Experiment name not manually set. Default to {experiment_name}."
        )
        sys.argv += [f"experiment_name={experiment_name}"]

    if (
        "--multirun" in sys.argv
        or "hydra.mode=MULTIRUN" in sys.argv
        or "-m" in sys.argv
    ):
        raise NotImplementedError("Hydra multi-run is not supported.")
    # non-multirun mode, add trial_name and hydra run dir
    if any("trial_name=" in x for x in sys.argv):
        trial_name = next(x for x in sys.argv if "trial_name=" in x).split("=")[
            1
        ]
    else:
        trial_name = (
            f"run{datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}"
        )
        sys.argv += [f"trial_name={trial_name}"]
    if "_" in trial_name:
        raise RuntimeError("trial_name should not contain `_`.")
    sys.argv += [
        f"hydra.run.dir={cluster_spec.fileroot}/logs/{getpass.getuser()}/"
        f"{experiment_name}/{trial_name}/hydra-outputs/"
    ]

    sys.argv.pop(1)

    func()


if __name__ == "__main__":
    main()



from typing import *
import os

import torch
import transformers

from realhf.api.core.model_api import ReaLModelConfig, register_hf_family
from realhf.base import constants


def convert_state_dict_llama(state_dict: Dict, config: ReaLModelConfig) -> Dict:

    new_state_dict = {}
    for k, v in state_dict.items():
        if k == "model.embed_tokens.weight":
            new_state_dict["0.wte.weight"] = v
        elif k == "lm_head.weight":
            new_state_dict[f"{config.n_layers + 1}.weight"] = v
        elif k == "model.norm.weight":
            new_state_dict[f"{config.n_layers}.ln_f.weight"] = v
        elif "inv_freq" in k:
            continue
        else:
            block_idx = int(k.split(".")[2])
            name = k.split(".", 3)[3]
            replace_pairs = [
                ("self_attn.", "attn."),
                ("post_attention_layernorm.", "mlp.ln."),
                ("input_layernorm.", "attn.c_attn.ln."),
                ("attn.o_proj.", "attn.c_proj."),
                ("q_proj.", "c_attn.q_attn."),
                ("k_proj.", "c_attn.k_attn."),
                ("v_proj.", "c_attn.v_attn."),
            ]
            for k1, k2 in replace_pairs:
                if k1 in name:
                    name = name.replace(k1, k2)
            new_state_dict[f"{block_idx + 1}.{name}"] = v

    if constants.use_te_impl():
        state_dict = new_state_dict
        new_state_dict = {}
        te_replace_pairs = [
            (".mlp.ln.weight", ".mlp.layer_norm_weight"),
            (".mlp.down_proj.weight", ".mlp.fc2_weight"),
        ]
        for k, v in state_dict.items():
            for k1, k2 in te_replace_pairs:
                if k1 in k:
                    k = k.replace(k1, k2)
            new_state_dict[k] = v

        # fuse gate && up weight
        for i in range(config.n_layers):
            gate_w = new_state_dict[f"{i+1}.mlp.gate_proj.weight"]
            upproj_w = new_state_dict[f"{i+1}.mlp.up_proj.weight"]
            w = torch.cat([gate_w, upproj_w], dim=0)
            new_state_dict[f"{i+1}.mlp.fc1_weight"] = w
            new_state_dict[f"{i+1}.mlp._extra_state"] = None
            new_state_dict.pop(f"{i+1}.mlp.gate_proj.weight")
            new_state_dict.pop(f"{i+1}.mlp.up_proj.weight")
    return new_state_dict


def to_llama_state_dict(
    state_dict: Dict[str, torch.Tensor], config: ReaLModelConfig
) -> Dict:
    if constants.use_te_impl():
        # remove all extra states
        keys = list(state_dict.keys())
        for k in keys:
            if k.endswith("_extra_state"):
                state_dict.pop(k)

        # split gate && up weight
        for i in range(config.n_layers):
            w = state_dict[f"{i + 1}.mlp.fc1_weight"]
            gate_w, upproj_w = w.split(
                (w.shape[0] // 2, w.shape[0] // 2), dim=0
            )
            state_dict[f"{i + 1}.mlp.gate_proj.weight"] = gate_w.contiguous()
            state_dict[f"{i + 1}.mlp.up_proj.weight"] = upproj_w.contiguous()
            state_dict.pop(f"{i + 1}.mlp.fc1_weight")

        # rename
        new_state_dict = {}
        te_replace_pairs = [
            (".mlp.layer_norm_weight", ".mlp.ln.weight"),
            (".mlp.fc2_weight", ".mlp.down_proj.weight"),
        ]
        for k, v in state_dict.items():
            for k1, k2 in te_replace_pairs:
                if k1 in k:
                    k = k.replace(k1, k2)
            new_state_dict[k] = v
        state_dict = new_state_dict

    _k = list(state_dict.keys())[0]
    device = state_dict[_k].device

    layer_indices = list(set([int(k.split(".")[0]) for k in state_dict.keys()]))

    new_sd = {}
    for i in layer_indices:
        if i == 0:
            new_sd["model.embed_tokens.weight"] = state_dict["0.wte.weight"]
        elif i == config.n_layers + 1:
            new_sd["lm_head.weight"] = state_dict[f"{i}.weight"]
        else:
            new_sd[f"model.layers.{i-1}.input_layernorm.weight"] = state_dict[
                f"{i}.attn.c_attn.ln.weight"
            ]
            new_sd[f"model.layers.{i-1}.mlp.down_proj.weight"] = state_dict[
                f"{i}.mlp.down_proj.weight"
            ]
            new_sd[f"model.layers.{i-1}.mlp.gate_proj.weight"] = state_dict[
                f"{i}.mlp.gate_proj.weight"
            ]
            new_sd[f"model.layers.{i-1}.mlp.up_proj.weight"] = state_dict[
                f"{i}.mlp.up_proj.weight"
            ]
            new_sd[f"model.layers.{i-1}.post_attention_layernorm.weight"] = (
                state_dict[f"{i}.mlp.ln.weight"]
            )
            new_sd[f"model.layers.{i-1}.self_attn.k_proj.weight"] = state_dict[
                f"{i}.attn.c_attn.k_attn.weight"
            ]
            new_sd[f"model.layers.{i-1}.self_attn.o_proj.weight"] = state_dict[
                f"{i}.attn.c_proj.weight"
            ]
            new_sd[f"model.layers.{i-1}.self_attn.q_proj.weight"] = state_dict[
                f"{i}.attn.c_attn.q_attn.weight"
            ]
            new_sd[f"model.layers.{i-1}.self_attn.v_proj.weight"] = state_dict[
                f"{i}.attn.c_attn.v_attn.weight"
            ]
            new_sd[f"model.layers.{i-1}.self_attn.rotary_emb.inv_freq"] = (
                1.0
                / (
                    config.rotary_base
                    ** (
                        torch.arange(
                            0,
                            config.head_dim,
                            2,
                            device=device,
                            dtype=torch.float32,
                        )
                        / config.head_dim
                    )
                )
            )
            if i == config.n_layers:
                new_sd["model.norm.weight"] = state_dict[f"{i}.ln_f.weight"]

    return new_sd


# param name is used to load directly from huggingface checkpoints
def llama_embedding_layer_names(config: ReaLModelConfig) -> List[str]:
    return ["model.embed_tokens.weight"]


def llama_transformer_block_param_name(
    config: ReaLModelConfig, idx: int
) -> List[str]:
    names = [
        f"model.layers.{idx}.input_layernorm.weight",
        f"model.layers.{idx}.mlp.down_proj.weight",
        f"model.layers.{idx}.mlp.gate_proj.weight",
        f"model.layers.{idx}.mlp.up_proj.weight",
        f"model.layers.{idx}.post_attention_layernorm.weight",
        f"model.layers.{idx}.self_attn.k_proj.weight",
        f"model.layers.{idx}.self_attn.o_proj.weight",
        f"model.layers.{idx}.self_attn.q_proj.weight",
        # f"model.layers.{idx}.self_attn.rotary_emb.inv_freq",
        f"model.layers.{idx}.self_attn.v_proj.weight",
    ]
    if idx == config.n_layers - 1:
        names += ["model.norm.weight"]
    return names


def llama_output_head_param_name(config: ReaLModelConfig) -> List[str]:
    return ["lm_head.weight"]


def convert_config_llama(
    hf_config: transformers.LlamaConfig,
) -> ReaLModelConfig:
    return ReaLModelConfig(
        n_layers=hf_config.num_hidden_layers,
        n_kv_heads=hf_config.num_key_value_heads,
        hidden_dim=hf_config.hidden_size,
        head_dim=hf_config.hidden_size // hf_config.num_attention_heads,
        intermediate_dim=hf_config.intermediate_size,
        vocab_size=hf_config.vocab_size,
        n_positions=hf_config.max_position_embeddings,
        embd_pdrop=0.0,
        attn_pdrop=(
            hf_config.attention_dropout
            if hasattr(hf_config, "attention_dropout")
            else 0.1
        ),
        layer_norm_epsilon=hf_config.rms_norm_eps,
        activation_function=hf_config.hidden_act,
        use_attention_bias=hf_config.attention_bias,
        scale_attn_by_inverse_layer_idx=False,
        layer_norm_type="rms",
        mlp_type="llama",
        apply_rotary=True,
        rotary_base=hf_config.rope_theta,
        rotary_interleaved=False,
        rotary_scaling=(
            None
            if hf_config.rope_scaling is None
            else hf_config.rope_scaling["factor"]
        ),
        rotary_scaling_type=(
            None
            if hf_config.rope_scaling is None
            else hf_config.rope_scaling["type"]
        ),
    )


def convert_config_back_llama(
    config: ReaLModelConfig,
) -> transformers.LlamaConfig:
    rope_scaling = {}
    if config.rotary_scaling is not None:
        rope_scaling["factor"] = config.rotary_scaling
    if config.rotary_scaling_type is not None:
        rope_scaling["type"] = config.rotary_scaling_type
    return transformers.LlamaConfig(
        vocab_size=config.vocab_size,
        hidden_size=config.hidden_dim,
        intermediate_size=config.intermediate_dim,
        num_hidden_layers=config.n_layers,
        num_key_value_heads=config.n_kv_heads,
        num_attention_heads=config.hidden_dim // config.head_dim,
        max_position_embeddings=config.n_positions,
        rms_norm_eps=config.layer_norm_epsilon,
        hidden_act=config.activation_function,
        attention_dropout=config.attn_pdrop,
        attention_bias=config.use_attention_bias,
        rope_theta=config.rotary_base,
        rope_scaling=rope_scaling if rope_scaling else None,
    )


for name in [
    "llama",
    "codellama",
    "deepseek",
]:
    register_hf_family(
        name=name,
        hf_cls_name="LlamaForCausalLM",
        config_from_hf_converter=convert_config_llama,
        config_to_hf_converter=convert_config_back_llama,
        sd_from_hf_converter=convert_state_dict_llama,
        sd_to_hf_converter=to_llama_state_dict,
        embedding_param_names=llama_embedding_layer_names,
        tblock_param_names=llama_transformer_block_param_name,
        head_param_names=llama_output_head_param_name,
    )



# Re-import these classes for clear documentation,
# otherwise the name will have a long prefix like
# realhf.api.quickstart.model.ModelTrainEvalConfig.
from .api.core.dfg import MFCDef
from .api.core.model_api import ModelInterface, ReaLModelConfig
from .api.quickstart.dataset import (
    PairedComparisonDatasetConfig,
    PromptAnswerDatasetConfig,
    PromptOnlyDatasetConfig,
)
from .api.quickstart.device_mesh import AllocationConfig
from .api.quickstart.model import (
    ModelTrainEvalConfig,
    OptimizerConfig,
    ParallelismConfig,
)
from .base.namedarray import NamedArray
from .experiments.common.common import CommonExperimentConfig
from .experiments.common.dpo_exp import DPOConfig
from .experiments.common.ppo_exp import PPOConfig, PPOHyperparameters
from .experiments.common.rw_exp import RWConfig
from .experiments.common.sft_exp import SFTConfig

__version__ = "0.1.0.post1"



from abc import ABC, abstractmethod

from realhf.impl.model.parallelism.pipeline_parallel.instruction import *


class PipeSchedule(ABC):
    """Directs the execution of a pipeline engine by generating sequences of
    :class:`PipeInstruction`.

    Schedules are generators that yield sequences of
    :class:`PipeInstruction` to process the micro-batches in one batch.
    Each yielded step is atomic in the sense that a barrier
    synchronization can be placed between successive steps without
    deadlock.

    Below is an example schedule that implements data parallelism with gradient accumulation:

    .. code-block:: python

        class DataParallelSchedule(PipeSchedule):
            def steps(self):
                for step_id in range(self.micro_batches):
                    cmds = [
                        LoadMicroBatch(buffer_id=0),
                        ForwardPass(buffer_id=0),
                        BackwardPass(buffer_id=0),
                    ]
                    if step_id == self.micro_batches - 1:
                        cmds.extend([
                            ReduceGrads(),
                            OptimizerStep(),
                        ])
                    yield cmds

            def num_pipe_buffers(self):
                return 1

    Args:
        micro_batches (int): The number of micro-batches that comprise a batch.
        stages (int): The number of pipeline stages.
        stage_id (int): The pipe stage that will execute the generated schedule.
    """

    def __init__(self, micro_batches, stages, stage_id):
        super().__init__()
        self.micro_batches = micro_batches
        self.stages = stages
        self.stage_id = stage_id
        self.prev_stage = self.stage_id - 1
        self.next_stage = self.stage_id + 1

        self.terminate_hooks = []
        self.current_micro_batch_id = 0

    def terminate(self):
        return self.terminate_hooks

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(micro_batches={self.micro_batches}, stages={self.stages}, stage_id={self.stage_id})"

    @abstractmethod
    def steps(self):
        """Yield a list of :class:`PipeInstruction` for each step in the schedule.

        .. note::
            Schedules must implement ``steps()`` to define the schedule.

        Returns:
            Instructions to be executed as one step of the pipeline
        """
        pass

    def num_pipe_buffers(self):
        """The number of pipeline buffers that will be used by this stage.

        .. note::
            Schedules should specialize ``num_pipe_buffers()`` for memory savings at scale.

        Returns:
            The number of buffers for the engine to allocate.
        """
        return self.micro_batches

    def _valid_micro_batch(self, micro_batch_id):
        return 0 <= micro_batch_id < self.micro_batches

    def _valid_stage(self, stage_id):
        return 0 <= stage_id < self.stages

    @property
    def stage(self):
        """Stage index used to configure this schedule."""
        return self.stage_id

    @property
    def num_stages(self):
        """The number of total pipeline stages used to configure this schedule."""
        return self.stages

    @property
    def num_micro_batches(self):
        """The number of total micro_batches used to configure this schedule."""
        return self.micro_batches

    @property
    def is_first_stage(self):
        """True if the configured ``stage_id`` is the first stage in the pipeline."""
        return self.stage_id == 0

    @property
    def is_last_stage(self):
        """True if the configured ``stage_id`` is the last stage in the pipeline."""
        return self.stage_id == self.stages - 1

    def _buffer_idx(self, micro_batch_id):
        """Map a micro-batch index to a pipeline buffer index.

        This method uses a cyclic allocation strategy.

        Args:
            micro_batch_id (int): The micro-batch index relative to the beginning of the schedule.

        Returns:
            int: The index of the buffer that should store data.
        """
        assert self._valid_micro_batch(micro_batch_id)
        return micro_batch_id % self.num_pipe_buffers()

    def __iter__(self):
        self.it = None
        return self

    def __next__(self):
        if self.it is None:
            self.it = self.steps()
        return next(self.it)


class InferenceSchedule(PipeSchedule):
    """A schedule for inferencing batches using pipeline parallelism."""

    def steps(self):
        """"""
        total_steps = self.micro_batches + self.stages - 1
        for step_id in range(total_steps):
            cmds = []
            micro_batch_id = step_id - self.stage_id

            if _is_even(self.stage_id):
                if self._valid_stage(self.next_stage):
                    if self._valid_micro_batch(micro_batch_id - 1):
                        cmds.append(
                            SendActivation(self.stage_id, micro_batch_id - 1)
                        )
                if self._valid_stage(self.prev_stage):
                    if self._valid_micro_batch(micro_batch_id):
                        cmds.append(
                            RecvActivation(self.stage_id, micro_batch_id)
                        )
            else:
                if self._valid_stage(self.prev_stage):
                    if self._valid_micro_batch(micro_batch_id):
                        cmds.append(
                            RecvActivation(self.stage_id, micro_batch_id)
                        )

                if self._valid_stage(self.next_stage):
                    if self._valid_micro_batch(micro_batch_id - 1):
                        cmds.append(
                            SendActivation(self.stage_id, micro_batch_id - 1)
                        )

            if self._valid_micro_batch(micro_batch_id):
                cmds.append(ForwardPass(self.stage_id, micro_batch_id))

            yield step_id, micro_batch_id, cmds

    def num_pipe_buffers(self):
        """Only two pipeline buffers are required for inferencing.

        Returns:
            ``2``
        """
        return 2


class GenerateSchedule(PipeSchedule):
    """A schedule for generate.
    Difference between this schedule and InferenceSchedule is that last stage will not load data,
    and the last stage will send the result to the first stage for the next generation round.
    """

    def __init__(self, micro_batches, stages, stage_id, max_new_tokens):
        super().__init__(micro_batches, stages, stage_id)
        self.prev_stage = self.prev_stage % self.stages
        self.next_stage = self.next_stage % self.stages
        self.max_new_tokens = max_new_tokens
        self.max_steps = (
            max_new_tokens * max(self.num_micro_batches, self.stages)
            + self.num_micro_batches
            - 1
        )  # a configurable upper bound

    def _valid_token_id(self, token_id):
        return token_id < self.max_new_tokens

    def steps(self):
        last_micro_batch_id = -1
        last_token_id = -1
        for step_id in range(self.max_steps):
            cmds = []
            micro_batch_id = (
                (step_id - self.stage_id)
                % max(self.num_micro_batches, self.stages)
                if step_id - self.stage_id >= 0
                else -1
            )  # micro batch id for current stage
            first_round = (
                step_id < self.num_micro_batches
            )  # whether it is the first round of generate
            last_stage_last_mbid = (
                (step_id - self.stages)
                % max(self.num_micro_batches, self.stages)
                if step_id >= self.stages
                else -1
            )
            # the micro_batch_id of the last stage on last step
            token_id = (step_id - self.stage_id) // max(
                self.num_micro_batches, self.stages
            )
            # token id in current round

            # TODO: from last stage to first stage, need one buffer for each microbatch?
            if _is_even(self.stage_id):
                if (
                    self._valid_micro_batch(last_micro_batch_id)
                    and self._valid_token_id(last_token_id)
                    and not self.is_last_stage
                ):
                    cmds.append(
                        SendActivation(
                            self.stage_id, last_micro_batch_id, step_id=token_id
                        )
                    )
                # intermediate stage recv
                if (
                    self._valid_micro_batch(micro_batch_id)
                    and self._valid_token_id(token_id)
                    and not self.is_first_stage
                ):
                    cmds.append(
                        RecvActivation(
                            self.stage_id, micro_batch_id, step_id=token_id
                        )
                    )
            else:
                # odd stage could not be first stage
                if self._valid_micro_batch(
                    micro_batch_id
                ) and self._valid_token_id(token_id):
                    cmds.append(
                        RecvActivation(
                            self.stage_id, micro_batch_id, step_id=token_id
                        )
                    )
                # last stage should not send activation except first stage requires
                if (
                    self._valid_micro_batch(last_micro_batch_id)
                    and self._valid_token_id(last_token_id)
                    and not self.is_last_stage
                ):
                    cmds.append(
                        SendActivation(
                            self.stage_id, last_micro_batch_id, step_id=token_id
                        )
                    )

            # last stage send next tokens when first stage requires.
            if (
                self.is_last_stage
                and self._valid_micro_batch(last_micro_batch_id)
                and self._valid_token_id(last_token_id)
            ):
                cmds.append(
                    SendNextTokens(
                        self.stage_id,
                        last_micro_batch_id,
                        step_id=last_token_id,
                    )
                )
            if self.is_first_stage and self._valid_micro_batch(
                last_stage_last_mbid
            ):
                cmds.append(
                    RecvNextTokens(
                        self.stage_id, last_stage_last_mbid, step_id=token_id
                    )
                )

            if self._valid_micro_batch(micro_batch_id) and self._valid_token_id(
                token_id
            ):
                cmds.append(
                    ForwardPass(self.stage_id, micro_batch_id, step_id=token_id)
                )

            last_micro_batch_id = micro_batch_id
            last_token_id = token_id
            yield step_id, micro_batch_id, cmds

    def num_pipe_buffers(self):
        """2 buffers for inter stage transfer (except last stage to first stage)
        self.num_micro_batches buffers for last stage to first stage transfer

        Returns:
            ``2 + self.num_micro_batches``
        """
        return 2  # + self.num_micro_batches


class TrainSchedule(PipeSchedule):
    """A schedule for training a batch using hybrid parallelism.

    Pipeline parallelism is extracted through gradient accumulation and thus
    convergence follows that of a data parallel approach with the same batch
    size.
    """

    def steps(self):
        """"""
        prev_micro_batch_id = -1
        total_steps = 2 * (self.micro_batches + self.stages - 1)
        for step_id in range(total_steps):
            # Map the step of the pipeline to the micro-batch id and also whether it is a
            # forward or backward pass step.
            micro_batch_id, is_forward = self._step_to_micro_batch(step_id)
            cmds = []

            # Exchange activations
            if is_forward:
                if self._valid_micro_batch(
                    prev_micro_batch_id
                ) and self._valid_stage(self.prev_stage):
                    cmds.append(
                        SendGrad(self.stage_id, prev_micro_batch_id, step_id=0)
                    )
                if self._valid_micro_batch(
                    micro_batch_id
                ) and self._valid_stage(self.prev_stage):
                    cmds.append(
                        RecvActivation(self.stage_id, micro_batch_id, step_id=0)
                    )
            else:
                if self._valid_micro_batch(
                    micro_batch_id
                ) and self._valid_stage(self.next_stage):
                    cmds.append(
                        RecvGrad(self.stage_id, micro_batch_id, step_id=0)
                    )
                if self._valid_micro_batch(
                    prev_micro_batch_id
                ) and self._valid_stage(self.next_stage):
                    cmds.append(
                        SendActivation(
                            self.stage_id, prev_micro_batch_id, step_id=0
                        )
                    )

            # Computation
            if self._valid_micro_batch(micro_batch_id):
                if is_forward:
                    cmds.append(
                        ForwardPass(self.stage_id, micro_batch_id, step_id=0)
                    )
                else:
                    cmds.append(
                        BackwardPass(self.stage_id, micro_batch_id, step_id=0)
                    )

            # Model step at the end of the batch
            if step_id == total_steps - 1:
                cmds.append(
                    ReduceGrads(self.stage_id, micro_batch_id, step_id=0)
                )
                cmds.append(
                    OptimizerStep(self.stage_id, micro_batch_id, step_id=0)
                )

            # Prepare state for next time
            prev_micro_batch_id = micro_batch_id
            # TODO: TEMP FOR DEBUG
            yield step_id, micro_batch_id, cmds

    def num_pipe_buffers(self):
        """Return the number of pipeline buffers required for this stage.

        This is equivalent to the maximum number of in-flight forward passes,
        since we need to remember the activations of forward passes in order
        to run backpropagation. For synchronous 1F1B, this is equivalent to
        the index difference between this stage and the last stage.
        """
        buffers = min(self.stages - self.stage_id, self.micro_batches)
        return max(2, buffers)

    def _step_to_micro_batch(self, step_id):
        if _is_even(step_id) and _is_even(self.stage_id):
            micro_batch_id = self._even_step_forward_id(step_id)
            is_forward = True

        elif _is_odd(step_id) and _is_odd(self.stage_id):
            micro_batch_id = self._odd_step_forward_id(step_id)
            is_forward = True

        elif _is_even(step_id) and _is_odd(self.stage_id):
            micro_batch_id = self._even_step_backward_id(step_id)
            is_forward = False

        elif _is_odd(step_id) and _is_even(self.stage_id):
            micro_batch_id = self._odd_step_backward_id(step_id)
            is_forward = False

        else:
            assert False

        return micro_batch_id, is_forward

    def _even_step_forward_id(self, step_id):
        base = step_id // 2
        micro_batch_id = int(base - self.stage_id // 2)
        return micro_batch_id

    def _odd_step_forward_id(self, step_id):
        base = (step_id - 1) // 2
        micro_batch_id = int(base - self.stage_id // 2)
        return micro_batch_id

    def _even_step_backward_id(self, step_id):
        base = step_id // 2
        micro_batch_id = int(base - self.stages + (self.stage_id + 1) // 2)
        return micro_batch_id

    def _odd_step_backward_id(self, step_id):
        base = ((step_id - 1) // 2) - self.stages + 1
        micro_batch_id = int(base + self.stage_id // 2)
        return micro_batch_id


class DataParallelSchedule(PipeSchedule):
    """An example schedule that trains using traditional data parallelism with gradient
    accumulation.
    """

    def steps(self):
        """"""
        for step_id in range(self.micro_batches):
            cmds = [
                ForwardPass(self.stage_id, step_id, step_id=0),
                BackwardPass(self.stage_id, step_id, step_id=0),
            ]
            if step_id == self.micro_batches - 1:
                cmds.extend(
                    [
                        ReduceGrads(self.stage_id, step_id, step_id=0),
                        OptimizerStep(self.stage_id, step_id, step_id=0),
                    ]
                )
            yield cmds

    def num_pipe_buffers(self):
        """Only one pipeline buffer needed."""
        return 1


def _is_even(x):
    return x % 2 == 0


def _is_odd(x):
    return x % 2 != 0



from typing import List, Optional
import dataclasses
import enum


class JobState(enum.Enum):
    NOT_FOUND = 0
    PENDING = 1
    RUNNING = 2
    COMPLETED = 3
    FAILED = 4
    CANCELLED = 5

    def active(self):
        return self == self.PENDING or self == self.RUNNING


class SchedulerError(Exception):
    pass


class JobException(Exception):

    def __init__(self, run_name, worker_type, host, reason: JobState):
        super().__init__(
            f"Job {run_name}:{worker_type} {reason} at node {host}"
        )
        self.run_name = run_name
        self.worker_type = worker_type
        self.host = host
        self.reason = reason


@dataclasses.dataclass
class JobInfo:
    name: str
    state: JobState
    host: str = (
        None  # The host on which the job is/was running. None if the job had not run.
    )
    submit_time: str = None
    start_time: str = None
    slurm_id: str = None  # Slurm only. The Slurm id of the job.


class SchedulerClient:

    def __init__(self, expr_name, trial_name):
        self.expr_name = expr_name
        self.trial_name = trial_name
        self.run_name = f"{expr_name}_{trial_name}"

    def submit(self, worker_type, cmd, **kwargs):
        """Submits a job to the scheduler. Raises exception if the job is already running.

        Args:
            worker_type: The worker type to be submitted. The job name is specified when initializing the client.
            cmd (str or List[str]): The command of this job. If this is str, the command is parsed by
                shell; otherwise it is executed directly.
        """
        raise NotImplementedError()

    def submit_array(self, worker_type, cmd, count, **kwargs):
        """Submits an array of jobs to the scheduler.

        Args:
            worker_type: The worker type to be submitted, shared by all jobs.
            cmd: Command template of the jobs that may contain an "{index}" format placeholder.
            count: Number of jobs. The indices of the jobs shall be 0..count-1.
        """
        for index in range(count):
            self.submit(
                worker_type + "_" + str(index),
                cmd.format(index=index, count=count),
                **kwargs,
            )

    def stop(self, job_name):
        """Stops a running job. Raises exception if there is no such job, but passes if the job has stopped
        either successfully or not.
        """
        raise NotImplementedError()

    def stop_all(self, signal=None):
        """Stops the whole job."""
        raise NotImplementedError()

    def find(self, job_name) -> Optional[JobInfo]:
        """Gets the status of a job of this job.

        Args:
            job_name: Name of the job.

        Returns:
            A JobInfo if the job is found, or None otherwise.
        """
        raise NotImplementedError()

    def find_all(self, job_name_regex=".*") -> List[JobInfo]:
        """Finds jobs.

        Args:
            job_name_regex: job name regex.

        Returns:
            A list of found JobInfo.
        """
        raise NotImplementedError()

    def wait(self, timeout=None, **kwargs):
        """Waits until all jobs submitted via this client instance finish."""
        raise NotImplementedError()


def remote_worker_cmd(expr_name, trial_name, debug, worker_type):
    # requires information in scheduler package
    return (
        f"python3 {'' if debug else '-O'} -m realhf.apps.remote worker -w {worker_type} "
        f"-e {expr_name} -f {trial_name} -i {{jobstep_id}} -g {{n_jobsteps}} -r {{worker_submission_index}} "
        f"-p {{wprocs_per_jobstep}} -j {{wprocs_in_job}} -o {{wproc_offset}}"
    )


def setup_cmd(expr_name, trial_name, debug):
    bash_cmd = (  # f"pip3 install -e $REAL_PACKAGE_PATH --no-build-isolation && "
        f"python3 {'' if debug else '-O'} -m realhf.apps.remote "
        f"reset_name_resolve -e {expr_name} -f {trial_name}"
    )
    # return f"bash -c \"{bash_cmd}\""
    return bash_cmd


def control_cmd(
    expr_name, trial_name, debug, ignore_worker_error, controller_type
):
    bash_cmd = (  # f"pip3 install -e $REAL_PACKAGE_PATH --no-build-isolation && "
        f"python3 {'' if debug else '-O'} -m realhf.apps.remote controller "
        f"-e {expr_name} -f {trial_name} "
        f"--{'ignore_worker_error' if ignore_worker_error else 'raise_worker_error'} "
        f"--type {controller_type}"
    )
    # return f"bash -c \"{bash_cmd}\""
    return bash_cmd


def ray_cluster_cmd(expr_name, trial_name, worker_type):
    flags = [f"-e {expr_name}", f"-f {trial_name}", f"-w {worker_type}"]
    return f"python3 -m realhf.apps.remote ray -i {{index}} -g {{count}} {' '.join(flags)}"


def make(mode, expr_name, trial_name, **kwargs) -> SchedulerClient:
    if mode == "slurm":
        from realhf.scheduler.slurm.client import SlurmSchedulerClient

        return SlurmSchedulerClient(expr_name, trial_name)
    elif mode == "local":
        from realhf.scheduler.local.client import LocalSchedulerClient

        return LocalSchedulerClient(expr_name, trial_name)
    else:
        raise NotImplementedError(f"Scheduler {mode} not found")



from typing import List, Optional, Tuple
import dataclasses
import functools

import torch


class LogitsWarper:
    """Abstract base class for all logit processors that can be applied during generation.

    Cloned from huggingface transformers/src/transformers/generation/logits_process.py,
    except that we can optionally change the logits inplace.
    """

    def __call__(
        self,
        input_ids: torch.LongTensor,
        logits: torch.FloatTensor,
        inplace: bool = False,
    ) -> torch.FloatTensor:
        raise NotImplementedError(
            f"{self.__class__} is an abstract class. Only classes inheriting this class can be called."
        )


@dataclasses.dataclass
class TemperatureLogitsWarper(LogitsWarper):
    temperature: float

    def __post_init__(self):
        if self.temperature > 1.0 or self.temperature < 0.0:
            raise ValueError("temperature has to be between 0 and 1")

    def __call__(
        self,
        _,
        logits: torch.FloatTensor,
        inplace: bool = False,
    ) -> Tuple[torch.FloatTensor, Optional[torch.FloatTensor]]:
        if inplace:
            logits.div_(self.temperature)
        else:
            logits = logits / self.temperature
        return logits / self.temperature


@dataclasses.dataclass
class TopPLogitsWarper(LogitsWarper):
    top_p: float
    filter_value: float = -float("Inf")
    min_tokens_to_keep: int = 1

    def __post_init__(self):
        self.top_p = top_p = float(self.top_p)
        if top_p < 0 or top_p > 1.0:
            raise ValueError(
                f"`top_p` has to be a float > 0 and < 1, but is {top_p}"
            )
        if not isinstance(self.min_tokens_to_keep, int) or (
            self.min_tokens_to_keep < 1
        ):
            raise ValueError(
                f"`min_tokens_to_keep` has to be a positive integer, "
                f"but is {self.min_tokens_to_keep}"
            )

    def __call__(
        self,
        _,
        logits: torch.FloatTensor,
        inplace: bool = False,
    ) -> Tuple[torch.FloatTensor, Optional[torch.FloatTensor]]:
        sorted_logits, sorted_indices = torch.sort(logits, descending=False)
        cumulative_probs = sorted_logits.softmax(dim=-1).cumsum(dim=-1)

        # Remove tokens with cumulative top_p above the threshold (token with 0 are kept)
        sorted_indices_to_remove = cumulative_probs <= (1 - self.top_p)
        # Keep at least min_tokens_to_keep
        sorted_indices_to_remove[..., -self.min_tokens_to_keep :] = 0

        # scatter sorted tensors to original indexing
        indices_to_remove = sorted_indices_to_remove.scatter(
            -1, sorted_indices, sorted_indices_to_remove
        )
        self.filter_value = torch.finfo(logits.dtype).min
        if inplace:
            logits.masked_fill_(indices_to_remove, self.filter_value)
        else:
            logits = logits.masked_fill(indices_to_remove, self.filter_value)
        return logits


@dataclasses.dataclass
class TopKLogitsWarper(LogitsWarper):
    top_k: int
    filter_value: float = -float("Inf")
    min_tokens_to_keep: int = 1

    def __post_init__(self):
        top_k = self.top_k
        min_tokens_to_keep = self.min_tokens_to_keep
        self.top_k = max(top_k, min_tokens_to_keep)
        if not isinstance(top_k, int) or top_k <= 0:
            raise ValueError(
                f"`top_k` has to be a strictly positive integer, but is {top_k}"
            )

    def __call__(
        self,
        _,
        logits: torch.FloatTensor,
        inplace: bool = False,
    ) -> Tuple[torch.FloatTensor, Optional[torch.FloatTensor]]:
        top_k = min(self.top_k, logits.size(-1))  # Safety check
        # Remove all tokens with a probability less than the last token of the top-k
        indices_to_remove = logits < torch.topk(logits, top_k)[0][..., -1, None]
        self.filter_value = torch.finfo(logits.dtype).min
        if inplace:
            logits.masked_fill_(indices_to_remove, self.filter_value)
        else:
            logits = logits.masked_fill(indices_to_remove, self.filter_value)
        return logits


@dataclasses.dataclass
class EpsilonLogitsWarper(LogitsWarper):
    epsilon: float
    filter_value: float = -float("Inf")
    min_tokens_to_keep: int = 1

    def __post_init__(self):
        self.epsilon = epsilon = float(self.epsilon)
        epsilon = float(epsilon)
        if epsilon <= 0 or epsilon >= 1:
            raise ValueError(
                f"`eta_cutoff` has to be a float > 0 and < 1, but is {epsilon}"
            )

        self.min_tokens_to_keep = min_tokens_to_keep = int(
            self.min_tokens_to_keep
        )
        if min_tokens_to_keep < 1:
            raise ValueError(
                f"`min_tokens_to_keep` has to be a strictly positive integer, but is {min_tokens_to_keep}"
            )

    def __call__(
        self,
        _,
        logits: torch.FloatTensor,
        inplace: bool = False,
    ) -> Tuple[torch.FloatTensor, Optional[torch.FloatTensor]]:
        # Calculate the adaptive cutoff
        probabilities = logits.softmax(dim=-1)
        entropy = torch.distributions.Categorical(logits).entropy()
        eta = torch.min(
            self.epsilon, torch.sqrt(self.epsilon) * torch.exp(-entropy)
        )[..., None]
        indices_to_remove = probabilities < eta

        # Keep the words with the 'min_tokens_to_keep'-highest probabilities
        top_k = min(self.min_tokens_to_keep, logits.size(-1))  # Safety check
        indices_to_remove = indices_to_remove & (
            logits < torch.topk(logits, top_k)[0][..., -1, None]
        )

        self.filter_value = torch.finfo(logits.dtype).min
        if inplace:
            logits.masked_fill_(indices_to_remove, self.filter_value)
        else:
            logits = logits.masked_fill(indices_to_remove, self.filter_value)
        return logits


def chained_logits_wraper(xs: List[LogitsWarper], inplace: bool = False):

    def foo(
        input_ids: torch.LongTensor,
        logits: torch.FloatTensor,
    ) -> torch.FloatTensor:
        for x in xs:
            logits = x(input_ids, logits, inplace)
        return logits

    return foo


def unioned_logits_wraper(xs: List[LogitsWarper], inplace: bool = False):

    def foo(
        input_ids: torch.LongTensor,
        logits: torch.FloatTensor,
    ) -> torch.FloatTensor:
        processed_logits = [x(input_ids, logits, inplace=False) for x in xs]
        masks = [logits != pl for pl in processed_logits]
        mask = functools.reduce(torch.logical_or, masks)
        if inplace:
            logits.masked_fill_(mask, torch.finfo(logits.dtype).min)
        else:
            logits = logits.masked_fill(mask, torch.finfo(logits.dtype).min)
        return logits

    return foo


def top_k_top_p_logits(
    logits: torch.Tensor,
    top_k=0,
    top_p=1.0,
    inplace: bool = False,
    ordered: bool = False,
) -> torch.FloatTensor:
    if top_p == 1.0 and top_k >= logits.shape[-1]:
        return logits
    if top_p == 1.0:
        return TopKLogitsWarper(top_k=top_k)(None, logits, inplace=inplace)
    if top_k >= logits.shape[-1]:
        return TopPLogitsWarper(top_p=top_p)(None, logits, inplace=inplace)
    warper_fn = unioned_logits_wraper if not ordered else chained_logits_wraper
    p = warper_fn(
        [TopKLogitsWarper(top_k=top_k), TopPLogitsWarper(top_p=top_p)],
        inplace=inplace,
    )
    return p(None, logits)



from typing import Callable, Dict, List, Optional
import itertools
import json

import numpy as np
import torch
import torch.utils.data

from realhf.api.core import data_api
from realhf.base.namedarray import NamedArray


class RewardModelingPairedDataset(torch.utils.data.Dataset):

    def __init__(
        self,
        util: data_api.DatasetUtility,
        max_length: int,
        max_pairs_per_prompt: int = 2,
        dataset_path: Optional[str] = None,
        dataset_builder: Optional[Callable[[], List[Dict]]] = None,
    ):
        """Dataset used for reward modeling. Each sample consists of a prompt, several positive answers, and several negative answers.

        Args:
            util (api.data.DatasetUtility): Dataset utility.
            max_length (int): The maximum sequence length. Sequences will be right-padded to this length
            dataset_path (Optional[str], optional): Path to the dataset json/jsonl file.
                The json/jsonl file should be a list of dictionary. Each element in the list should have
                the key "prompt", "pos_answers", and "neg_answers". Each "pos_answer" must correspond to
                one and only one "neg_answer" (i.e., they are one-to-one pairs). Defaults to None.
            dataset_builder (Optional[Callable[[], List[Dict]]], optional): Alternative to dataset_path.
                A callable that returns a list of dictionary. Defaults to None.
        """
        self.util = util
        seed = self.util.seed
        tokenizer = self.util.tokenizer

        self.max_pairs_per_prompt = max_pairs_per_prompt

        self.rng = np.random.RandomState(seed=seed)
        data = data_api.load_shuffle_split_dataset(
            util, dataset_path, dataset_builder
        )

        prompts = [x["prompt"] for x in data]

        pos_answers = [
            [x["prompt"] + c + tokenizer.eos_token for c in x["pos_answers"]]
            for x in data
        ]
        neg_answers = [
            [x["prompt"] + c + tokenizer.eos_token for c in x["neg_answers"]]
            for x in data
        ]

        for a, b in zip(pos_answers, neg_answers):
            if len(a) != len(b):
                raise RuntimeError(
                    "pos_answers and neg_answers must be one-to-one pairs."
                )
            if len(a) == 0:
                raise RuntimeError(
                    "pos_answers and neg_answers must be non-empty."
                )

        group_sizes = [len(x) for x in pos_answers]

        self.prompt_tokens = tokenizer(
            prompts,
            max_length=max_length,
            truncation=True,
            padding=False,
            return_length=True,
        )
        _pos_answer_tokens = tokenizer(
            list(itertools.chain.from_iterable(pos_answers)),
            max_length=max_length,
            padding=False,
            truncation=True,
            return_length=True,
        )
        _neg_answer_tokens = tokenizer(
            list(itertools.chain.from_iterable(neg_answers)),
            max_length=max_length,
            padding=False,
            truncation=True,
            return_length=True,
        )

        pos_answer_tokens = []
        neg_answer_tokens = []
        offset = 0
        for g in group_sizes:
            pos_answer_tokens.append(
                {
                    k: v[offset : offset + g]
                    for k, v in _pos_answer_tokens.items()
                }
            )
            neg_answer_tokens.append(
                {
                    k: v[offset : offset + g]
                    for k, v in _neg_answer_tokens.items()
                }
            )
            offset += g

        self.pos_answer_tokens: List[List[int]] = pos_answer_tokens
        self.neg_answer_tokens: List[List[int]] = neg_answer_tokens
        assert (
            len(self.prompt_tokens["input_ids"])
            == len(self.pos_answer_tokens)
            == len(self.neg_answer_tokens)
        )

    def __len__(self):
        return len(self.pos_answer_tokens)

    def __getitem__(self, idx):
        prompt_len = self.prompt_tokens["length"][idx]
        n_pairs_this_prompt = len(self.pos_answer_tokens[idx]["input_ids"])
        group_size = min(self.max_pairs_per_prompt, n_pairs_this_prompt)
        pair_indices = self.rng.choice(
            n_pairs_this_prompt, group_size, replace=False
        )

        packed_input_ids = []
        for i in pair_indices:
            packed_input_ids += self.pos_answer_tokens[idx]["input_ids"][i]
            packed_input_ids += self.neg_answer_tokens[idx]["input_ids"][i]

        pos_input_lens = [
            len(self.pos_answer_tokens[idx]["input_ids"][i])
            for i in pair_indices
        ]
        neg_input_lens = [
            len(self.neg_answer_tokens[idx]["input_ids"][i])
            for i in pair_indices
        ]

        input_lens = [x + y for x, y in zip(pos_input_lens, neg_input_lens)]
        assert sum(input_lens) == len(packed_input_ids)

        x = NamedArray(
            packed_input_ids=torch.tensor(packed_input_ids, dtype=torch.long),
            pos_input_lens=torch.tensor(pos_input_lens, dtype=torch.int32),
            group_factor=torch.tensor(
                [1 / group_size for _ in range(group_size)], dtype=torch.float32
            ),
            prompt_lens=torch.tensor(
                [prompt_len for _ in range(group_size)], dtype=torch.int32
            ),
        )
        assert (
            x["pos_input_lens"].shape
            == x["group_factor"].shape
            == x["prompt_lens"].shape
        )
        assert x["pos_input_lens"].shape[0] == len(input_lens)
        x.register_metadata(seqlens=input_lens)
        return x


data_api.register_dataset("rw_pair", RewardModelingPairedDataset)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here.
import pathlib
import sys

sys.path.insert(
    0, (pathlib.Path(__file__).parents[2] / "realhf").resolve().as_posix()
)

project = "ReaL"
copyright = "2024, Wei Fu & Zhiyu Mei"
author = "Wei Fu & Zhiyu Mei"
release = "0.1.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_nefertiti"
html_static_path = ["_static"]

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.githubpages",
]


def setup(app):
    app.add_css_file("custom.css")



from typing import *
import sys


class PipeInstruction:
    """Base class for all instructions to be executed by the pipeline engine.

    All keyword arguments are stored as members similar to a ``namedtuple``. These are
    then accessible to the :class:`PipeEngine` during execution.

    Each instruction has following attributes:
    1. stage_id: pipeline stage that the instruction should be executed by.
    2. micro_batch_id: the ID of data micro batch that this instruction should be executed on.
    3. step_id: usually used by generation schedules, identical to generation token id.
    4. priority: priority of the instruction, higher priority instructions should be scheduled first in
                 the dynamic schedule.
    5. deps: list of instructions that this instruction depends on.
    6. bind: Instruction that this instruction is binded with. Binded instructions are send/recv instructions
             that should appear in pair, belong to different (adjancent) stages and have the same micro_batch_id.
             Two binded instructions should have the same dependency list. This feature is used to avoid NCCL
             communication deadlock.

    """

    def __init__(
        self,
        stage_id: int,
        micro_batch_id: int,
        deps: List["PipeInstruction"] = [],
        bind: List["PipeInstruction"] = [],
        step_id: int = 0,
        schedule_id: int = 0,
    ):
        self.stage_id = stage_id
        self.micro_batch_id = micro_batch_id
        self.deps = deps
        self.bind = bind
        self.name = self.__class__.__name__
        self.step_id = step_id
        self.args = (stage_id, micro_batch_id, step_id)
        self.__str_encode = (
            f"{self.name};{self.stage_id};{self.micro_batch_id};{self.step_id}"
        )
        self.schedule_id = schedule_id

    def __repr__(self):
        return f"{self.name}{self.args}"
        # return call_to_str(self.name, self.args, self.kwargs)

    def __eq__(self, other: "PipeInstruction"):
        return (
            self.stage_id == other.stage_id
            and self.micro_batch_id == other.micro_batch_id
            and self.step_id == other.step_id
            and self.name == other.name
        )

    def __lt__(self, other: "PipeInstruction"):
        # order by stage_id, micro_batch_id, step_id
        # used to sort finded results in InstructionSet
        return (
            self.stage_id < other.stage_id
            or (
                self.stage_id == other.stage_id
                and self.micro_batch_id < other.micro_batch_id
            )
            or (
                self.stage_id == other.stage_id
                and self.micro_batch_id == other.micro_batch_id
                and self.step_id < other.step_id
            )
        )

    def encode_str(self) -> str:
        return self.__str_encode

    def encode(self):
        return self.encode_str().encode(encoding="utf-8")

    @classmethod
    def decode(cls, encoded: Union[bytes, str]) -> "PipeInstruction":
        if isinstance(encoded, bytes):
            s = encoded.decode(encoding="utf-8")
        else:
            s = encoded
        cls_name, stage_id, micro_batch_id, step_id = s.split(";")
        cls_ = getattr(sys.modules[__name__], cls_name)
        return cls_(
            stage_id=int(stage_id),
            micro_batch_id=int(micro_batch_id),
            step_id=int(step_id),
        )


def decode_stage_by_encoded(s: str):
    return int(s.split(";")[1])


class OptimizerStep(PipeInstruction):
    """Performs one step with the optimizer and zeros gradients.

    .. note:: Should be issued after :class:`ReduceGrads` and :class:`ReduceTiedGrads`.

    .. note:: Can be a synchronization point among data-parallel ranks.
    """

    pass


class ReduceGrads(PipeInstruction):
    """Reduce the computed gradients among data-parallel processes within the stage."""

    pass


# Compute
class ForwardPass(PipeInstruction):
    """Compute a forward pass."""

    pass


class BackwardPass(PipeInstruction):
    """Compute a backward pass and accumulate gradients."""

    pass


# Communication
class SendActivation(PipeInstruction):
    """Send activations to the next stage in the pipeline."""

    pass


class RecvActivation(PipeInstruction):
    """Receive activations from the previous stage in the pipeline."""

    pass


class SendGrad(PipeInstruction):
    """Send computed gradients to the previous pipeline stage.
    with respect to the received activations
    """

    pass


class RecvGrad(PipeInstruction):
    """Receive computed gradients the next pipeline stage."""

    pass


# generation
class SendNextTokens(PipeInstruction):
    """In GenerateSchedule, send next tokens to the first stage. Only available in the last stage."""

    pass


class RecvNextTokens(PipeInstruction):
    """In GenerateSchedule, recv next tokens from the last stage. Only available in the first stage."""

    pass



from typing import Dict, List, Tuple

import numpy as np


def shape_leq(shape1: Tuple, shape2: Tuple) -> bool:
    assert len(shape1) == len(shape2)
    return all(x1 <= x2 for x1, x2 in zip(shape1, shape2))


def shape_union(*shapes: List[Tuple]) -> Tuple:
    if len(shapes) == 1:
        return shapes[0]
    for s in shapes:
        assert len(s) == len(shapes[0])
    return tuple(max(*dims) for dims in zip(*shapes))


def split_to_shapes(x: np.ndarray, shapes: Dict, axis: int = -1):
    """Split an array and reshape to desired shapes.

    Args:
        x (np.ndarray): The array to be splitted
        shapes (Dict): Dict of shapes (tuples) specifying how to split.
        axis (int): Split dimension.

    Returns:
        List: Splitted observations.
    """
    axis = len(x.shape) + axis if axis < 0 else axis
    split_lengths = [np.prod(shape) for shape in shapes.values()]
    assert x.shape[axis] == sum(split_lengths)
    accum_split_lengths = [
        sum(split_lengths[:i]) for i in range(1, len(split_lengths))
    ]
    splitted_x = np.split(x, accum_split_lengths, axis)
    return {
        k: x.reshape(*x.shape[:axis], *shape, *x.shape[axis + 1 :])
        for x, (k, shape) in zip(splitted_x, shapes.items())
    }



# Copied from https://github.com/microsoft/DeepSpeed
from collections import defaultdict
from typing import Any, List, Optional, Tuple

import torch

import realhf.base.logging as logging
import realhf.impl.model.parallelism.pipeline_parallel.p2p as p2p

logger = logging.getLogger("tensor_utils")


def get_shape(tensor):
    return tensor.shape if torch.is_tensor(tensor) else None


def print_data_shapes(name, rank, mbid, x, ys):
    if rank == 0:
        logger.debug(f"{name}: rank {rank} mbid {mbid}")
        logger.debug(
            f"shapes: x.pp_input {get_shape(x.pp_input)}, x.pp_output {get_shape(x.pp_output)},"
            f" x.cu_seqlens {get_shape(x.cu_seqlens)}"
        )
        for i, y in enumerate(ys):
            logger.debug(
                f"shapes: ys[{i}].input_ids {get_shape(y.packed_input_ids)}, "
                f"ys[{i}].k_cache {get_shape(y.k_cache)}, ys[{i}].v_cache {get_shape(y.v_cache)}, "
                f"ys[{i}].cache_seqlens {get_shape(y.cache_seqlens)}"
            )


class TensorBuffer:
    # could store both tensors and other data

    def __init__(self):
        self.tensors = defaultdict(dict)

    def put(self, name: str, mbid: int, x: torch.Tensor):
        self.tensors[name][mbid] = x

    def alloc(
        self,
        name: str,
        mbid: int,
        shape: Tuple[int],
        dtype: torch.dtype,
        device: torch.device,
        require_grads: bool = False,
    ):
        self.tensors[name][mbid] = torch.zeros(
            shape, dtype=dtype, device=device, requires_grad=require_grads
        )
        return self.tensors[name][mbid]

    def get(
        self,
        name: str,
        mbid: int,
        remove: bool = False,
        raise_error: bool = True,
    ):
        try:
            if remove:
                return self.tensors[name].pop(mbid)
            else:
                return self.tensors[name][mbid]
        except KeyError as e:
            if raise_error:
                raise e
            else:
                return None

    def remove(
        self, name: str, mbid: Optional[int] = None, check_exists: bool = False
    ):
        try:
            if mbid is None:
                del self.tensors[name]
            else:
                self.tensors[name].pop(mbid)
        except KeyError:
            if not check_exists:
                return
            raise KeyError(
                f"TensorBuffer.remove: key {name} mbid {mbid} not found"
            )

    def check_name(self, name: str):
        return name in self.tensors

    def check_mbid(self, name: str, mbid: int):
        if name not in self.tensors:
            return False
        return mbid in self.tensors[name]

    def clear(self):
        self.tensors = defaultdict(dict)



from dataclasses import asdict
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple
import copy
import dataclasses
import enum
import getpass
import json
import os
import time
import traceback

import ray
import ray.util.queue as rq

from realhf.base import logging, name_resolve, names
from realhf.base.cluster import spec as cluster_spec
from realhf.system import load_worker, worker_base, worker_control, WORKER_TYPES
from realhf.system.worker_base import WorkerServerStatus as Wss
import realhf.api.core.system_api as system_api

CONNECTION_RETRY_AFTER_SECONDS = 360

logger = logging.getLogger("controller")


@dataclasses.dataclass
class TrialStatus:
    experiment_name: str
    trial_name: str
    running_workers: Dict[str, List[str]] = dataclasses.field(
        default_factory=dict
    )


@dataclasses.dataclass
class TrialHistory:
    experiment_name: str
    trial_name: str
    age_days: int


class ControllerExitStatus(enum.Enum):
    SUCCESS = 0
    TIMEOUT = 1
    INTERRUPTED = 9
    FAIL = 101
    LOST = 102
    UNKNOWN = 404


class Controller:

    def __init__(
        self, experiment_name, trial_name, panel: worker_base.WorkerControlPanel
    ):
        assert "_" not in experiment_name, (
            f"_ not allowed in experiment_name (args: -e) "
            f"{experiment_name}, use '-' instead."
        )
        assert (
            "_" not in trial_name
        ), f"_ not allowed in trial_name (args: -f) {trial_name}, use '-' instead."
        self.experiment_name = experiment_name
        self.trial_name = trial_name

        logger.info("Experiment: %s %s", self.experiment_name, self.trial_name)

        self.__control = panel
        self.json_config_file_path = f"{cluster_spec.fileroot}/logs/{getpass.getuser()}/{self.experiment_name}_{self.trial_name}"

    def reconnect(self):
        """Automatically reconnect to workers. And list all jobs to scheduler."""
        self.__control.auto_connect()

    def start(
        self, experiment: system_api.Experiment, ignore_worker_error=False
    ):
        if ignore_worker_error:
            check_worker_status = ()
            remove_worker_status = (
                Wss.COMPLETED,
                Wss.ERROR,
                Wss.LOST,
                Wss.UNKNOWN,
            )
        else:
            check_worker_status = (Wss.ERROR, Wss.LOST, Wss.UNKNOWN)
            remove_worker_status = (Wss.COMPLETED,)

        scheduling: system_api.ExperimentScheduling = (
            experiment.scheduling_setup()
        )
        setup = experiment.initial_setup()
        setup.set_worker_information(
            experiment_name=self.experiment_name, trial_name=self.trial_name
        )

        if setup.config is not None:
            os.makedirs(self.json_config_file_path, exist_ok=True)
            with open(
                os.path.join(self.json_config_file_path, "config.json"), "w"
            ) as f:
                json.dump(asdict(setup.config), f, indent=4)

        # Scheduling and connecting to workers.
        # TODO for heterogeneous workers of the same type k, list scheduling[k] and setup[k] should match.
        workers_configs = [
            (k, getattr(setup, k), getattr(scheduling, k)) for k in WORKER_TYPES
        ]

        for name, config, schedule in workers_configs:
            count = (
                sum([s.count for s in schedule])
                if isinstance(schedule, list)
                else schedule.count
            )
            if len(config) != count:
                logger.error(
                    "Scheduling and config mismatch, interrupting all workers."
                )
                self.interrupt()
                raise IndexError(
                    f"Configuration has {len(config)} {name}, {count} scheduled."
                )
            logger.info(f"Configuration has {len(config)} {name}.")

        name_resolve.add(
            names.trial_registry(self.experiment_name, self.trial_name),
            value=datetime.now().strftime("%Y%m%d"),
            delete_on_exit=False,
            replace=True,
        )
        name_resolve.add(
            names.worker_status(
                experiment_name=self.experiment_name,
                trial_name=self.trial_name,
                worker_name="ctl",
            ),
            value="READY",
            delete_on_exit=True,
        )

        while True:
            try:
                logger.info("Connecting to workers...")
                self.__control.connect(
                    [
                        self.__control.name(name, i)
                        for name, cfgs, _ in workers_configs
                        for i in range(len(cfgs))
                    ],
                    progress=True,
                    timeout=CONNECTION_RETRY_AFTER_SECONDS,
                    raises_timeout_error=True,
                )
                break

            except TimeoutError:
                logger.info("Connecting to workers timeout. Retrying...")
            except KeyboardInterrupt as e:
                logger.info("Interrupted by user. Stopping all and exiting...")
                raise e

        name_resolve.delete(
            names.worker_status(
                experiment_name=self.experiment_name,
                trial_name=self.trial_name,
                worker_name="ctl",
            )
        )

        # Configure workers.
        try:
            for name, cfgs, _ in workers_configs:
                logger.info(f"Configuring Workers: {name}...")
                self.__control.group_request(
                    "configure",
                    worker_names=[
                        self.__control.name(name, i) for i in range(len(cfgs))
                    ],
                    worker_kwargs=[dict(config=cfg) for cfg in cfgs],
                    progress=True,
                )
        except Exception as e:
            logger.error(f"Configuring Failed: {e}. Exiting Workers.")
            logger.error(traceback.format_exc())
            self.interrupt(wait_timeout=120)
            raise e

        logger.info("Start workers...")
        self.__control.group_request("start")
        logger.info("Started.")
        try:
            self.wait(
                timeout=None,
                check_status=check_worker_status,
                remove_status=remove_worker_status,
            )
        except worker_base.WorkerException as e:
            logger.error(e)
            self.interrupt(wait_timeout=30)
        except KeyboardInterrupt:
            logger.info("Interrupted.")
            self.interrupt(wait_timeout=30)

    def wait(
        self,
        timeout: Optional[int],
        check_status: Tuple[Wss, ...],
        remove_status: Tuple[Wss, ...],
    ):
        deadline = None if timeout is None else time.time() + timeout
        left = set(self.__control.worker_names)
        num_jobs_left = len(left)
        logger.info(f"Waiting for {num_jobs_left} jobs.")
        current_status = {
            name: Wss.UNKNOWN for name in self.__control.worker_names
        }
        while len(left) > 0:
            logger.debug(
                f"JOBS LEFT: {[str(len([l for l in left if job_type in l])) + ' ' + job_type for job_type in set([job_id.split('/')[0] for job_id in left])]}"
            )
            if len(left) < num_jobs_left:
                num_jobs_left = len(left)
                logger.info(f"Waiting for {num_jobs_left} jobs.")
            if deadline is not None and time.time() > deadline:
                raise TimeoutError(
                    f"Timeout waiting for {self.experiment_name, self.trial_name}: {', '.join(sorted(left))}"
                )
            for worker_name, worker_status in self.__control.pulse().items():
                if worker_status in check_status:
                    raise worker_base.WorkerException(
                        worker_name, worker_status, "experiment is running."
                    )
                if worker_status in remove_status:
                    if worker_name in current_status:
                        logger.debug(
                            f"Worker {worker_name} is {worker_status}. Removed from waiting list."
                        )
                        current_status.pop(worker_name)
                    else:
                        pass
                else:
                    if current_status.get(worker_name, None) != worker_status:
                        current_status.update({worker_name: worker_status})
                        logger.debug(
                            f"Update worker status: {worker_name} -> {worker_status}"
                        )

            left = set(current_status.keys())
            time.sleep(10)

    def stop(self):
        """Stop the experiment.
        Note:
            This method assumes that the controller and scheduler is connected to the correct workers. To ensure this,
            call controller.reconnect before your call controller.stop.
        """
        raise NotImplementedError()

    def interrupt(self, wait_timeout=120):
        """Interrupt the experiment."""
        logger.info("Interrupting experiment")
        self.__control.group_request("interrupt", wait_response=False)
        try:
            self.wait(
                timeout=wait_timeout,
                check_status=(),
                remove_status=(
                    Wss.ERROR,
                    Wss.LOST,
                    Wss.COMPLETED,
                    Wss.INTERRUPTED,
                ),
            )
        except TimeoutError:
            raise RuntimeError(
                f"Fail to interrupt workers, timeout={wait_timeout}."
            )


def run_ray_worker(
    worker_type,
    idx,
    experiment_name,
    trial_name,
    comm: Tuple[rq.Queue, rq.Queue],
):
    worker_name = f"{worker_type}/{idx}"
    server = worker_control.make_server(
        "ray",
        worker_name=worker_name,
        experiment_name=experiment_name,
        trial_name=trial_name,
        comm=comm,
    )
    worker = load_worker(worker_type)(server=server)
    try:
        worker.run()
    except Exception as e:
        logging.error("Worker %s failed with exception: %s", worker_name, e)
        logging.error(traceback.format_exc())
        raise e


class RayController:
    """A controller that uses Ray to manage workers.

    It uses the basic Controller to configure workers.
    Besides, it launchs all remote workers using Ray,
    instead of submitting them to the scheduelr.
    """

    def __init__(self, experiment_name, trial_name, local_mode: bool):
        # base controller will be lazier initialized when launching workers.
        self.__experiment_name = experiment_name
        self.__trial_name = trial_name
        self.__base_controller = None

        self.__workers_reply_comm = None
        self.__workers_request_comm = None
        self.__workers_ref = None

        self.__local_mode = local_mode

    def _launch_workers(
        self, workers_configs: List[Tuple[str, List, system_api.TasksGroup]]
    ):
        # Launch remote workers.
        logger.info("Launching remote workers using Ray...")
        self.__workers_ref: Dict[str, ray.ObjectRef] = {}
        self.__workers_request_comm: Dict[str, rq.Queue] = dict()
        self.__workers_reply_comm: Dict[str, rq.Queue] = dict()
        for worker_type, config, schedule in workers_configs:
            count = len(config)
            all_schedules: List[system_api.TasksGroup] = []
            if isinstance(schedule, List):
                for s in schedule:
                    for _ in range(s.count):
                        s_ = copy.deepcopy(s)
                        s_.count = 1
                        all_schedules.append(s_)
            else:
                for _ in range(schedule.count):
                    s_ = copy.deepcopy(schedule)
                    s_.count = 1
                    all_schedules.append(s_)
            assert len(all_schedules) == len(config)
            comms = [
                (rq.Queue(maxsize=8), rq.Queue(maxsize=8))
                for _ in all_schedules
            ]
            jobs = [
                ray.remote(
                    num_cpus=sch.scheduling.cpu,
                    num_gpus=sch.scheduling.gpu,
                    memory=sch.scheduling.mem,
                    name=f"{worker_type}/{idx}",
                )(run_ray_worker).remote(
                    worker_type,
                    idx,
                    self.__experiment_name,
                    self.__trial_name,
                    comm,
                )
                for idx, (comm, sch) in enumerate(zip(comms, all_schedules))
            ]
            for idx, (job, c) in enumerate(zip(jobs, comms)):
                name = f"{worker_type}/{idx}"
                self.__workers_ref[name] = job
                self.__workers_request_comm[name] = c[0]
                self.__workers_reply_comm[name] = c[1]
            logger.info(f"Launched {count} {worker_type}.")

        panel = worker_control.make_control(
            "ray",
            self.__experiment_name,
            self.__trial_name,
            request_comms=self.__workers_request_comm,
            reply_comms=self.__workers_reply_comm,
        )
        self.__base_controller = Controller(
            self.__experiment_name, self.__trial_name, panel
        )
        logger.info("All Ray workers are lauched.")

    def start(
        self, experiment: system_api.Experiment, ignore_worker_error=False
    ):
        scheduling: system_api.ExperimentScheduling = (
            experiment.scheduling_setup()
        )
        setup = experiment.initial_setup()
        setup.set_worker_information(
            experiment_name=self.__experiment_name, trial_name=self.__trial_name
        )
        workers_configs = [
            (k, getattr(setup, k), getattr(scheduling, k)) for k in WORKER_TYPES
        ]
        workers_configs: List[Tuple[str, List, system_api.TasksGroup]]

        if self.__local_mode:
            ray.init()
        else:
            for name, config, schedule in workers_configs:
                count = (
                    sum([s.count for s in schedule])
                    if isinstance(schedule, list)
                    else schedule.count
                )
                if len(config) != count:
                    logger.error(
                        "Scheduling and config mismatch, interrupting all workers."
                    )
                    raise IndexError(
                        f"Configuration has {len(config)} {name}, {count} scheduled."
                    )
                for idx in range(count):
                    try:
                        name_resolve.wait(
                            names.ray_cluster(
                                self.__experiment_name,
                                self.__trial_name,
                                f"{name}/{idx}",
                            ),
                            timeout=300,
                        )
                    except TimeoutError:
                        raise RuntimeError(
                            f"Timeout waiting for Ray cluster node {name}/{idx} to start."
                        )
            logger.info("Ray cluster started.")

            try:
                ray_head_addr = name_resolve.wait(
                    names.ray_cluster(
                        self.__experiment_name, self.__trial_name, "address"
                    ),
                    timeout=300,
                )
            except TimeoutError:
                raise RuntimeError(
                    "Timeout waiting for ray cluster head address."
                )
            ray.init(address=ray_head_addr)

        logger.info("Ray initialized! Ready to run workers.")

        try:
            self._launch_workers(workers_configs)
            self.__base_controller.start(experiment, ignore_worker_error)
        except Exception as e:
            self.shutdown()

    def shutdown(self):
        ray_exiting_name = names.ray_cluster(
            self.__experiment_name, self.__trial_name, "exiting"
        )
        name_resolve.add(ray_exiting_name, value="1", delete_on_exit=True)
        del self.__workers_reply_comm
        del self.__workers_request_comm
        del self.__workers_ref
        ray.shutdown()



from typing import Any, Dict, List, Optional, Tuple, Union
import pickle
import socket

import ray.util.queue as rq
import zmq

from realhf.base import logging
from realhf.system.worker_base import WorkerServerStatus
import realhf.system.worker_base as worker_base

logger = logging.getLogger("worker-control")
WORKER_WAIT_FOR_CONTROLLER_SECONDS = 3600
WORKER_JOB_STATUS_LINGER_SECONDS = 60


class ZmqTaskQueue(worker_base.WorkerServerTaskQueue):

    def __init__(self, port=0):
        self.__context = zmq.Context()
        self.__socket = self.__context.socket(zmq.REP)
        host_ip = socket.gethostbyname(socket.gethostname())
        if port == 0:
            self.__port = self.__socket.bind_to_random_port(f"tcp://{host_ip}")
        else:
            self.__socket.bind(f"tcp://{host_ip}:{port}")
            self.__port = port

    def __del__(self):
        self.__socket.close()

    @property
    def port(self):
        return self.__port

    def try_get_request(self) -> Tuple[str, Dict[str, Any]]:
        try:
            data = self.__socket.recv(zmq.NOBLOCK)
        except zmq.ZMQError:
            # Currently no request in the queue.
            raise worker_base.NoRequstForWorker()
        return pickle.loads(data)

    def respond(self, response):
        self.__socket.send(pickle.dumps(response))


class RayTaskQueue(worker_base.WorkerServerTaskQueue):

    def __init__(self, comm: Tuple[rq.Queue, rq.Queue]):
        recv_queue, send_queue = comm
        self.__recv_queue = recv_queue
        self.__send_queue = send_queue

    def try_get_request(self) -> Tuple[str, Dict[str, Any]]:
        try:
            command, kwargs = self.__recv_queue.get_nowait()
        except rq.Empty:
            # Currently no request in the queue.
            raise worker_base.NoRequstForWorker()
        return command, kwargs

    def respond(self, response):
        self.__send_queue.put(response)


class ZmqRequester(worker_base.WorkerControlPanelRequester):

    class ZmqFuture(worker_base.WorkerControlPanelRequester.Future):
        # Every ZmqFuture connect one socket, close after returning results.
        def __init__(
            self,
            payload,
            context: zmq.Context,
            address,
            worker_name,
            wait_response=True,
        ):
            self.__worker_name = worker_name
            self.__socket = context.socket(zmq.REQ)
            self.__socket.setsockopt(zmq.LINGER, 0)
            self.__socket.connect(f"tcp://{address}")
            self.__socket.send(payload, flags=zmq.NOBLOCK)
            if not wait_response:
                self.__socket.close()

        def result(self, timeout=None):
            if timeout is not None:
                self.__socket.RCVTIMEO = int(timeout * 1000)
            else:
                self.__socket.RCVTIMEO = int(1e9)
            try:
                r = pickle.loads(self.__socket.recv())
            except zmq.error.Again as e:
                raise TimeoutError(
                    f"Waiting for RPC server response timeout: {e}"
                )
            if isinstance(r, Exception):
                logger.error(f"Error configuring worker {self.__worker_name}")
                raise r
            self.__socket.close()
            return r

    def __init__(self):
        self.__context = zmq.Context()
        self.__context.set(zmq.MAX_SOCKETS, 20480)

    def async_request(
        self, worker_name, address, command, wait_response=True, **kwargs
    ):
        return self.ZmqFuture(
            pickle.dumps((command, kwargs)),
            self.__context,
            address,
            worker_name,
            wait_response=wait_response,
        )


class RayRequester(worker_base.WorkerControlPanelRequester):

    class RayQueueFuture(worker_base.WorkerControlPanelRequester.Future):

        def __init__(self, worker_name: str, queue: rq.Queue):
            self.__queue = queue
            self.__worker_name = worker_name

        def result(self, timeout=None):
            try:
                return self.__queue.get(timeout=timeout)
            except rq.Empty:
                raise TimeoutError(
                    f"Waiting for Ray worker {self.__worker_name} response timeout."
                )
            except Exception as e:
                raise RuntimeError(
                    f"Error waiting for Ray queue future {self.__worker_name}."
                ) from e

    def __init__(
        self,
        request_comms: Dict[str, rq.Queue],
        reply_comms: Dict[str, rq.Queue],
    ):
        self.__request_comms: Dict[str, rq.Queue] = request_comms
        self.__reply_comms: Dict[str, rq.Queue] = reply_comms

    def async_request(self, worker_name, _, command, __, **kwargs):
        request_queue = self.__request_comms[worker_name]
        request_queue.put((command, kwargs))
        reply_queue = self.__reply_comms[worker_name]
        return self.RayQueueFuture(worker_name, reply_queue)


def make_server(type_, worker_name, experiment_name, trial_name, **kwargs):
    if type_ == "zmq":
        q = ZmqTaskQueue(**kwargs)
    elif type_ == "ray":
        q = RayTaskQueue(**kwargs)
    else:
        raise NotImplementedError(type_)
    return worker_base.WorkerServer(worker_name, experiment_name, trial_name, q)


def make_control(type_, experiment_name, trial_name, **kwargs):
    if type_ == "zmq":
        requester = ZmqRequester(**kwargs)
    elif type_ == "ray":
        requester = RayRequester(**kwargs)
    else:
        raise NotImplementedError(type_)
    return worker_base.WorkerControlPanel(
        experiment_name, trial_name, requester
    )



import realhf.api.core.model_api as model_api


def check_is_realhf_native_impl(_cls):
    return _cls.__module__.startswith("realhf")


def check_is_realhf_native_model_interface(name):
    import realhf.impl.model.interface.dpo_interface
    import realhf.impl.model.interface.ppo_interface
    import realhf.impl.model.interface.rw_interface
    import realhf.impl.model.interface.sft_interface

    _cls = model_api.ALL_INTERFACE_CLASSES.get(name)
    return _cls and check_is_realhf_native_impl(_cls)
